from collections.abc import Awaitable, Callable
from functools import partial, partialmethod
from types import CodeType, CoroutineType, FunctionType, GeneratorType, MethodWrapperType
from typing import Any, TypeGuard

from .alias import (
    C_CALLABLES,
    BoundClassMethodC,
    BoundDunderMethodC,
    BoundMethodOrFuncC,
    BoundMethodT,
    FunctionT,
    ModuleT,
    UnboundDunderMethodC,
    UnboundMethodC,
)
from .code import FLAGS

__all__ = [
    'is_generator_function',
    'has_coroutine_mark',
    'is_coroutine_function',
    'is_async_generator_function',
    'is_awaitable',
    'is_function_with_flag',
    'unwrap_partial',
    'unwrap_partialmethod',
    'callable_location',
    'callable_to_pyfunc',
    'callable_module_and_name',
    'resolve_bound_method',
    'resolve_callable',
    'resolve_type_factory',
    'wrap_function_as',
    'replace_code_source_location',
]


# adapted from inspect:


def is_generator_function(func: Callable) -> bool:
    """Return True if the object is a user-defined generator function."""
    return is_function_with_flag(func, FLAGS.GENERATOR)


def has_coroutine_mark(fun: Any) -> bool:
    return hasattr(fun, '_is_coroutine_marker')


def is_coroutine_function(func: Callable) -> bool:
    """Return True if the object is a coroutine function.

    Coroutine functions are normally defined with "async def" syntax, but may
    be marked via markcoroutinefunction.
    """
    return is_function_with_flag(func, FLAGS.COROUTINE) or has_coroutine_mark(func)


def is_async_generator_function(func: Callable) -> bool:
    """Return true if the object is an asynchronous generator function.

    Asynchronous generator functions are defined with "async def"
    syntax and have "yield" expressions in their body.
    """
    return is_function_with_flag(func, FLAGS.ASYNC_GENERATOR)


def is_awaitable(obj: Any) -> TypeGuard[Awaitable]:
    """Return true if object can be passed to an ``await`` expression."""
    if isinstance(obj, CoroutineType):
        return True
    if isinstance(obj, GeneratorType) and bool(obj.gi_code.co_flags & FLAGS.ITERABLE_COROUTINE):
        return True
    return isinstance(obj, Awaitable)


def is_function_with_flag(func: Callable, flag: int) -> bool:
    """Return True if `func` is a function (or a method or functools.partial
    wrapper wrapping a function or a functools.partialmethod wrapping a
    function) whose code object has the given `flag` set in its flags."""
    f = unwrap_partialmethod(func)
    while isinstance(f, BoundMethodT):
        f = f.__func__
    f = unwrap_partial(f)
    if isinstance(f, FunctionT):
        return bool(f.__code__.co_flags & flag)
    return False


# adapted from functools:


def unwrap_partial(func: Callable) -> Callable:
    while isinstance(func, partial):
        func = func.func
    return func


def unwrap_partialmethod(func: Callable) -> Callable:
    prev = None
    while func is not prev:
        prev = func
        while isinstance(getattr(func, "__partialmethod__", None), partialmethod):
            func = func.__partialmethod__  # type: ignore
        while isinstance(func, partialmethod):
            func = getattr(func, 'func')
        func = unwrap_partial(func)
    return func


def callable_module_and_name(func: Callable) -> tuple[str, str, str]:
    if isinstance(func, FunctionT):
        mod = func.__module__
        qua = func.__qualname__
        nam = func.__name__

    elif isinstance(func, (UnboundDunderMethodC, BoundClassMethodC, UnboundMethodC)):
        cls = func.__objclass__
        mod = cls.__module__
        qua = func.__qualname__
        nam = func.__name__

    elif isinstance(func, BoundDunderMethodC):
        cls = func.__self__.__class__
        mod = cls.__module__
        qua = func.__qualname__
        nam = func.__name__

    elif isinstance(func, BoundMethodT):
        cls = func.__self__.__class__
        fun = func.__func__
        mod = getattr(fun, '__module__', '') or cls.__module__
        qua = fun.__qualname__
        nam = fun.__name__

    elif isinstance(func, BoundMethodOrFuncC):
        mod_or_self = func.__self__
        if isinstance(mod_or_self, ModuleT):
            mod = mod_or_self.__name__
        else:
            mod = mod_or_self.__module__
        qua = func.__qualname__
        nam = func.__name__

    else:
        mod = getattr(func, '__module__', '')
        qua = getattr(func, '__qualname__', '')
        nam = getattr(func, '__name__', '')

    mod = mod if isinstance(mod, str) else ''
    qua = qua if isinstance(qua, str) else ''
    nam = nam if isinstance(nam, str) else ''
    qua = qua or nam
    nam = nam or qua
    return mod, qua, nam


def callable_location(obj: object) -> str:
    while wrapped := getattr(obj, '__wrapped__', None):
        if not callable(wrapped):
            break
        obj = wrapped
    func = callable_to_pyfunc(obj)
    code = func.__code__
    return f'{code.co_filename}:{code.co_firstlineno}'


def callable_to_pyfunc(obj: object) -> FunctionType:
    assert callable(obj), f"expected a callable object, got {obj!r}"
    func, _ = resolve_callable(obj)
    return func


# MethodDescriptorType      as UnboundMethodC,
# WrapperDescriptorType     as UnboundDunderMethodC,
# MethodWrapperType         as BoundDunderMethodC,
# ClassMethodDescriptorType as BoundClassMethodC,
# BuiltinFunctionType       as BoundMethodOrFuncC,
# GetSetDescriptorType      as MutablePropertyC,
# MemberDescriptorType      as SlotPropertyC,


def resolve_bound_method(func: Callable) -> FunctionType:
    """Tries to resolve a bound method to an underlying FunctionType."""

    if isinstance(func, BoundMethodT):
        return func.__func__  # type: ignore

    if isinstance(func, BoundDunderMethodC):
        return resolve_shed_method(func.__self__.__class__, func.__name__)[0]

    if isinstance(func, BoundMethodOrFuncC) and not isinstance(func.__self__, ModuleT):
        return resolve_shed_method(func.__self__.__class__, func.__name__)[0]

    raise TypeError(f"{func} is not a bound method.")


def resolve_callable(func: Callable, n: int = 0) -> tuple[FunctionType, int]:
    """Tries to resolve a callable to an underlying FunctionType, giving the
    number of arguments on the left that should be ignored because they will
    be auto-provided by a binding mechanism we traversed through."""

    if isinstance(func, FunctionType):
        return func, n

    if isinstance(func, BoundMethodT):
        return resolve_callable(func.__func__, n + 1)

    # e.g. bytes.hex
    if isinstance(func, C_CALLABLES):
        return resolve_c_callable(func, n)

    # if the mystery object has a `__call__` method, we resolve to it:
    if callable(call_func := getattr(func, '__call__')):
        return resolve_callable(call_func, n)

    assert callable(func), f'callable({func}) is False'

    raise ValueError(f'cannot resolve {func} to an underlying python function')


def resolve_c_callable(func: Callable, n: int = 0) -> tuple[FunctionType, int]:
    """Tries to resolve a builtin callable object to an underlying FunctionType."""

    from .shed.load import get_shed_function

    # examples: list.append, dict.get, str.__hash__
    if isinstance(func, (UnboundMethodC, UnboundDunderMethodC)):
        return resolve_shed_method(func.__objclass__, func.__name__, n)

    # examples: bytes.fromhex
    if isinstance(func, BoundClassMethodC):
        return resolve_shed_method(func.__objclass__, func.__name__, n)

    # examples: ''.__hash__
    if isinstance(func, BoundDunderMethodC):
        return resolve_shed_method(func.__self__.__class__, func.__name__, n + 1)

    # examples: math.sin, builtins.len (functions) or [].append, {}.get (methods)
    if isinstance(func, BoundMethodOrFuncC):
        # builtins.len
        if isinstance(func.__self__, ModuleT):
            return get_shed_function(func.__self__.__name__, func.__name__), n

        # [].append
        return resolve_shed_method(func.__self__.__class__, func.__name__, n)

    raise TypeError(f"{func} is not one of {C_CALLABLES})")


def resolve_shed_method(cls: type, name: str, n: int = 0) -> tuple[FunctionType, int]:
    from .shed.load import get_shed_method

    shed_fn = get_shed_method(cls, name)
    if isinstance(shed_fn, staticmethod):
        return shed_fn.__func__, n - 1  # type: ignore
    elif isinstance(shed_fn, classmethod):
        return shed_fn.__func__, n  # type: ignore
    else:
        return shed_fn, n


def resolve_type_factory(cls: type) -> tuple[FunctionType, int]:
    """Tries to resolve to the FunctionType that `cls()` ends up calling."""

    # built-in classes resolve to their `__new__` in the type shed
    if cls.__flags__ & 256:
        return resolve_shed_method(cls, '__new__', 0)

    cls_call = cls.__call__

    # does this bind to `type`?
    if isinstance(cls_call, MethodWrapperType) and cls_call.__objclass__ == type:
        # if so, we know that this will call cls.__new__ followed by cls.__init__
        return resolve_callable(cls.__init__, 1)

    return resolve_callable(cls_call, 0)


TYPE_CALL = type.__dict__['__call__']  # type: ignore


def get_closure_value(func: FunctionT, pos: int) -> object:
    assert type(func) is FunctionT, f'{func} is not FunctionT'
    closure = func.__closure__
    assert closure is not None and len(closure) >= pos, f'{func} closure is too short'
    return closure[pos].cell_contents


FUNCTION_COPY_FIELDS = (
    '__module__',
    '__name__',
    '__qualname__',
    '__doc__',
    '__annotations__',
    '__type_params__',
    '__defaults__',
    '__kwdefaults__',
)


def wrap_function_as(
    old_fn: FunctionType, new_fn: FunctionType, hide_source: bool = True
) -> FunctionType:
    from copy import copy

    new_fn = copy(new_fn)
    for attr in FUNCTION_COPY_FIELDS:
        if hasattr(old_fn, attr):
            setattr(new_fn, attr, getattr(old_fn, attr))
    if hide_source:
        new_fn.__code__ = replace_code_source_location(old_fn.__code__, new_fn.__code__)
    return new_fn


# this linetable sends all traces for the old function to the first line of the new function:
INFINITE_LINETABLE = b'\x87\x00\x00' * 256


def replace_code_source_location(old_code: CodeType, new_code: CodeType) -> CodeType:
    return new_code.replace(
        co_firstlineno=old_code.co_firstlineno,
        co_linetable=INFINITE_LINETABLE,
        co_filename=old_code.co_filename,
        co_name=old_code.co_name,
    )



import agentica_internal.testing.examples.modules.mixin.a as a
import agentica_internal.testing.examples.modules.mixin.b as b
import agentica_internal.testing.examples.modules.mixin.c as c
import agentica_internal.testing.examples.modules.mixin.post as post


def verify_mixin():
    """basic mixin: child module overrides are called by inherited functions"""
    print(f'{a.x=}')
    print(f'{a.foo()=}')
    print(f'{a.bar()=}')
    print(f'{b.foo()=}')
    print(f'{b.bar()=}')
    print(f'{b.x=}')
    assert a.x == 3 * 10
    assert a.foo() == 'a.foo'
    assert a.bar() == ('a.bar', 'a.foo')
    assert b.foo() == 'b.foo'
    assert b.bar() == ('a.bar', 'b.foo')
    assert b.x == 2 * 10


def verify_closure_rewriting():
    """mixin rewrites functions captured in decorator closures"""
    # b.decorated_caller should use b.helper and b.foo
    print(f'{b.decorated_caller()=}')
    assert b.decorated_caller() == ['b.helper', 'b.foo']

    # c.decorated_caller should use c.helper and c.foo (both overridden)
    print(f'{c.decorated_caller()=}')
    assert c.decorated_caller() == ['c.helper', 'c.foo']

    # basic mixin still works for c
    print(f'{c.foo()=}')
    print(f'{c.bar()=}')
    assert c.foo() == 'c.foo'
    assert c.bar() == ('a.bar', 'c.foo')


def verify_init_module_globals():
    """
    init_module hook writes to child module's globals, not parent's.

    Pattern:
    1. Parent defines a module-level dict (_REGISTRY)
    2. @init_module hook populates it by scanning for is_*/has_* functions
    3. A function (get_registry) returns this dict
    4. Child modules should see their own functions in the dict
    """
    # a should only have is_feature_a
    a_reg = a.get_registry()
    print(f'{a_reg=}')
    assert 'is_feature_a' in a_reg
    assert 'is_feature_b' not in a_reg
    assert 'is_feature_c' not in a_reg

    # b should have is_feature_a (inherited) AND is_feature_b (its own)
    b_reg = b.get_registry()
    print(f'{b_reg=}')
    assert 'is_feature_a' in b_reg
    assert 'is_feature_b' in b_reg
    assert 'is_feature_c' not in b_reg

    # c should have all three
    c_reg = c.get_registry()
    print(f'{c_reg=}')
    assert 'is_feature_a' in c_reg
    assert 'is_feature_b' in c_reg
    assert 'is_feature_c' in c_reg


def verify_finalize():
    """finalize runs code after module is loaded"""
    print(f'{post.x=}')
    assert post.x == 20
    print(f'{post.y=}')
    assert post.y == 30


if __name__ == '__main__':
    verify_mixin()
    verify_closure_rewriting()
    verify_init_module_globals()
    verify_finalize()
    print('All mixin tests passed!')



"""
This a hand-curated typeshed file for methods in the `builtins` module.
"""

from collections.abc import Iterator
from typing import Any

####################################################################################################

__all__ = [
    '___init__',
    '___new__',
    '___getitem__',
    '___setitem__',
    '___delitem__',
    '___missing__',
    '___iter__',
    '___reversed__',
    '___contains__',
    '___class_getitem__',
    '___call__',
    '___instancecheck__',
    '___subclasscheck__',
    '___index__',
    '___bool__',
    '___getattr__',
    '___setattr__',
    '___delattr__',
    '___get__',
    '___set__',
    '___delete__',
    '___dir__',
    '___len__',
    '___hash__',
    '___repr__',
    '___str__',
    '___lt__',
    '___le__',
    '___eq__',
    '___ne__',
    '___ge__',
    '___gt__',
    '___not__',
    '___truth__',
    '___is___',
    '___is_not__',
    '___abs__',
    '___add__',
    '___and___',
    '___floordiv__',
    '___index__',
    '___inv__',
    '___invert__',
    '___lshift__',
    '___mod__',
    '___mul__',
    '___matmul__',
    '___neg__',
    '___or___',
    '___pos__',
    '___pow__',
    '___rshift__',
    '___sub__',
    '___truediv__',
    '___xor__',
    '___iadd__',
    '___iand__',
    '___iconcat__',
    '___ifloordiv__',
    '___ilshift__',
    '___imod__',
    '___imul__',
    '___imatmul__',
    '___ior__',
    '___ipow__',
    '___irshift__',
    '___isub__',
    '___itruediv__',
    '___ixor__',
]

####################################################################################################


def ___init__(self, *args: Any, **kwargs) -> None: ...
def ___new__(cls, *args: Any, **kwargs) -> Any: ...
def ___getitem__(self, key: Any) -> Any: ...
def ___setitem__(self, key: Any, value: Any) -> Any: ...
def ___delitem__(self, key: Any) -> None: ...
def ___missing__(self, key: Any) -> Any: ...
def ___iter__(self) -> Iterator: ...
def ___reversed__(self) -> Iterator: ...
def ___contains__(self, item: Any) -> bool: ...
def ___class_getitem__(cls, obj: Any) -> Any: ...
def ___call__(self, *args: Any, **kwargs: Any) -> Any: ...
def ___instancecheck__(self, instance: object) -> bool: ...
def ___subclasscheck__(self, subclass: type) -> bool: ...
def ___index__(self) -> int: ...
def ___bool__(self) -> bool: ...
def ___getattr__(self, attr: str) -> Any: ...
def ___setattr__(self, attr: str, value: Any) -> None: ...
def ___delattr__(self, attr: str) -> None: ...
def ___get__(self, instance: Any, owner: Any) -> Any: ...
def ___set__(self, instance: Any, value: Any) -> None: ...
def ___delete__(self, instance: Any) -> None: ...
def ___dir__(self) -> list[str]: ...
def ___len__(self) -> int: ...
def ___hash__(self) -> int: ...
def ___repr__(self) -> str: ...
def ___str__(self) -> str: ...


def ___lt__(self, b: Any, /) -> Any: ...
def ___le__(self, b: Any, /) -> Any: ...
def ___eq__(self, b: Any, /) -> Any: ...
def ___ne__(self, b: Any, /) -> Any: ...
def ___ge__(self, b: Any, /) -> Any: ...
def ___gt__(self, b: Any, /) -> Any: ...
def ___not__(self, /) -> bool: ...
def ___truth__(self, /) -> bool: ...
def ___is___(self, b: Any, /) -> bool: ...
def ___is_not__(self, b: Any, /) -> bool: ...
def ___abs__(self, /) -> Any: ...
def ___add__(self, b: Any, /) -> Any: ...
def ___and___(self, b: Any, /) -> Any: ...
def ___floordiv__(self, b: Any, /) -> Any: ...
def ___index__(self, /) -> int: ...
def ___inv__(self, /) -> Any: ...
def ___invert__(self, /) -> Any: ...
def ___lshift__(self, b: Any, /) -> Any: ...
def ___mod__(self, b: Any, /) -> Any: ...
def ___mul__(self, b: Any, /) -> Any: ...
def ___matmul__(self, b: Any, /) -> Any: ...
def ___neg__(self, /) -> Any: ...
def ___or___(self, b: Any, /) -> Any: ...
def ___pos__(self, /) -> Any: ...
def ___pow__(self, b: Any, /) -> Any: ...
def ___rshift__(self, b: Any, /) -> Any: ...
def ___sub__(self, b: Any, /) -> Any: ...
def ___truediv__(self, b: Any, /) -> Any: ...
def ___xor__(self, b: Any, /) -> Any: ...
def ___iadd__(self, b: Any, /) -> Any: ...
def ___iand__(self, b: Any, /) -> Any: ...
def ___iconcat__(self, b: Any, /) -> Any: ...
def ___ifloordiv__(self, b: Any, /) -> Any: ...
def ___ilshift__(self, b: Any, /) -> Any: ...
def ___imod__(self, b: Any, /) -> Any: ...
def ___imul__(self, b: Any, /) -> Any: ...
def ___imatmul__(self, b: Any, /) -> Any: ...
def ___ior__(self, b: Any, /) -> Any: ...
def ___ipow__(self, b: Any, /) -> Any: ...
def ___irshift__(self, b: Any, /) -> Any: ...
def ___isub__(self, b: Any, /) -> Any: ...
def ___itruediv__(self, b: Any, /) -> Any: ...
def ___ixor__(self, b: Any, /) -> Any: ...


# def __length_hint__(self, default: int = 0, /) -> int: ...
# def __concat__(a: Sequence, b: Sequence, /) -> Sequence: ...
# def __contains__(a: Container, b: Any, /) -> bool: ...
# def __delitem__(a: MutableSequence | MutableMapping, b: Any, /) -> None: ...
# def __getitem__(a: Sequence, b: slice, /) -> Sequence: ...
# def __indexOf__(a: Iterable, b: Any, /) -> int: ...
# def __setitem__(a: MutableSequence, b: Any, c: Any, /) -> None: ...



from enum import StrEnum
from typing import Literal

from msgspec import Struct, field, json

from ..core.datetime import time_now_utc
from ..core.utils import subclasses_dict
from ..internal_errors import *  # noqa: F403
from ..internal_errors import AgenticaError, ServerError
from ..session_manager_messages import PromptTemplate

MultiplexInvocationEventType = Literal['ENTER', 'EXIT', 'ERROR']

MultiplexErrorType = AgenticaError

MULTIPLEX_ERROR_CLASSES: dict[str, type[Exception]] = subclasses_dict(AgenticaError)
MULTIPLEX_ERROR_CLASSES['RuntimeError'] = RuntimeError  # for unknown errors
MULTIPLEX_ERROR_NAMES = list(MULTIPLEX_ERROR_CLASSES.keys())

MultiplexErrorName = StrEnum("MultiplexErrorName", MULTIPLEX_ERROR_NAMES)


def make_timestamp() -> str:
    return time_now_utc()


timestamp_field = field(default_factory=make_timestamp)


####################################################################################################

# Triggering events from the client


class MultiplexInvokeMessage(Struct, tag="invoke"):
    match_id: str
    uid: str
    warp_locals_payload: bytes
    streaming: bool
    parent_uid: str | None = None
    parent_iid: str | None = None
    prompt: str | PromptTemplate | None = None
    timestamp: str = timestamp_field


class MultiplexCancelMessage(Struct, tag="cancel"):
    uid: str
    iid: str
    timestamp: str = timestamp_field


# Generic carrier
class MultiplexDataMessage(Struct, tag="data"):
    uid: str
    iid: str
    data: bytes
    timestamp: str = timestamp_field


####################################################################################################


# Response from the server
class MultiplexNewIIDResponse(Struct, tag="new_iid"):
    uid: str
    iid: str
    match_id: str
    timestamp: str = timestamp_field


class MultiplexErrorMessage(Struct, tag="error"):
    iid: str
    error_name: str  # this could be InternalErrorName, but this lets us send other exceptions
    error_message: str | None = None
    timestamp: str = timestamp_field
    uid: str | None = None
    session_id: str | None = None
    session_manager_id: str | None = None

    @classmethod
    def from_error(
        cls,
        iid: str,
        err: AgenticaError,
        uid: str | None = None,
        session_id: str | None = None,
        session_manager_id: str | None = None,
    ) -> "MultiplexErrorMessage":
        err.__module__ = ""
        return cls(
            iid=iid,
            error_name=err.__class__.__name__,
            error_message=str(err),
            uid=uid,
            session_id=session_id,
            session_manager_id=session_manager_id,
        )

    def to_exception(self) -> Exception:
        error_name = self.error_name
        error_msg = self.error_message
        error_cls = MULTIPLEX_ERROR_CLASSES.get(error_name)
        if error_cls is None:
            error_cls = ServerError
            error_msg = f'{error_name}: {error_msg}' if error_msg else error_name

        exc = error_cls(error_msg or 'Unknown error')

        if isinstance(exc, AgenticaError):
            exc.uid = self.uid
            exc.iid = self.iid
            exc.session_id = self.session_id
            exc.session_manager_id = self.session_manager_id
            exc.error_timestamp = self.timestamp

        return exc


class MultiplexInvocationEventMessage(Struct, tag="invocation"):
    uid: str
    iid: str
    event: MultiplexInvocationEventType
    timestamp: str = timestamp_field


####################################################################################################

MultiplexMessage = (
    MultiplexDataMessage
    | MultiplexErrorMessage
    | MultiplexInvocationEventMessage
    | MultiplexNewIIDResponse
    | MultiplexCancelMessage
    | MultiplexInvokeMessage
)

MultiplexClientMessage = MultiplexInvokeMessage | MultiplexDataMessage | MultiplexCancelMessage

MultiplexClientInstanceMessage = MultiplexDataMessage | MultiplexCancelMessage

MultiplexServerMessage = (
    MultiplexNewIIDResponse
    | MultiplexErrorMessage
    | MultiplexDataMessage
    | MultiplexInvocationEventMessage
)

MultiplexServerInstanceMessage = (
    MultiplexErrorMessage | MultiplexDataMessage | MultiplexInvocationEventMessage
)

####################################################################################################

_enc_json = json.Encoder()
_dec_json = json.Decoder(MultiplexMessage)


def multiplex_from_json(json_bytes: bytes) -> MultiplexMessage:
    return _dec_json.decode(json_bytes)


def multiplex_to_json(message: MultiplexMessage) -> bytes:
    return _enc_json.encode(message)



from collections import defaultdict
from collections.abc import Callable
from types import CellType, FrameType, FunctionType, ModuleType
from typing import Any, Concatenate

from agentica_internal.core.collections import chaindict

__all__ = [
    'mixin',
    'init_module',
    'bases',
    'module',
    'finalize',
]

MIXIN_MODULE_INIT_HOOK_NAME = '___init_submodule___'
MIXIN_MODULE_BASES = '___mixing_module_bases___'
MIXIN_MODULE_SELF = '___mixing_module_self___'


def module(caller_globals: dict[str, Any] | None = None) -> ModuleType:
    """get the current module"""
    import sys

    if caller_globals is None:
        caller_globals = sys._getframe(1).f_globals
    dct = chaindict.top(caller_globals) or caller_globals
    if MIXIN_MODULE_SELF in dct:
        return dct[MIXIN_MODULE_SELF]
    raise ValueError('No module found')


def mixin(
    mod: ModuleType,
    *,
    include: tuple[str, ...] | None = None,
    exclude: tuple[str, ...] = (),
    **init_module_kwargs: Any,
) -> None:
    """
    Mixin a module's functions into the caller's namespace.

    Notably, mixed-in functions will have the caller's module as their `__globals__`
    global namespace, meaning any overridden/redefined functions in the caller's module
    will be called by the implicitly mixed-in functions, instead of the non-overridden ones
    in the original module.

    Essentially, this gives us singleton module-level function/method inheritance.

    Example:
    ```python a.py
    def foo():
        print('a.foo')

    def bar():
        foo()
        print('a.bar')
    ```
    ```python b.py
    import a
    from a import *
    mixin(a)

    def foo():
        print('b.foo')

    if __name__ == '__main__':
        bar()
        # b.foo
        # a.bar
    ```
    Without `mixin`, `bar()` would print `a.foo` and `a.bar`.
    """
    import inspect
    import sys

    caller_frame = sys._getframe(1)
    caller_globals: dict[str, Any] = caller_frame.f_globals
    caller_mod_name: str = caller_globals['__name__']
    sub_mod = sys.modules[caller_mod_name]  # the new module that is inheriting from `mod`
    src = mod.__dict__

    _attach_self(sub_mod)
    _attach_self(mod)

    # populate bases
    _add_base(sub_mod, mod)

    if include is None:
        names = [name for name in src if name not in exclude]
    else:
        names = include

    _sentinel = object()

    def _transform(_):
        """transform the module's functions (inheritance) and run the init_module hook"""
        for name in names:
            if (override := caller_globals.get(name, _sentinel)) is not _sentinel:
                if origin := getattr(override, '__module__', None):
                    # if an object is redefined, i.e. it comes from the same module as the caller,
                    # then we don't want to override it
                    if origin == caller_mod_name:
                        # this won't happen if `mixin()` happens at the top of the module anyway
                        continue
                else:
                    # don't override if object does not have __module__
                    continue

            obj = src[name]

            if origin := getattr(obj, '__module__', None):
                if origin != mod.__name__:
                    # anything that has __module__ and is not actually from the module
                    # we're mixing in, should not be overridden.
                    continue

            if inspect.isfunction(obj):
                new_func = _rewrite_function_globals(obj, caller_globals, caller_mod_name)
                caller_globals[name] = new_func
            else:
                caller_globals[name] = obj

        # equivalent to `__init_subclass__`, but for when a module is mixin'd
        if init_module_hook := mod.__dict__.get(MIXIN_MODULE_INIT_HOOK_NAME):
            # rewrite the hook's globals so `global X` statements affect sub_mod, not mod
            rewritten_hook = _rewrite_function_globals(
                init_module_hook, caller_globals, caller_mod_name
            )
            rewritten_hook(sub_mod, **init_module_kwargs)

    finalize(_transform, caller_frame)


def init_module[**Args](
    init: Callable[Concatenate[ModuleType, Args], None],
) -> Callable[Concatenate[ModuleType, Args], None]:
    """decorator: run function as an initialization hook on module when it is mixed-in"""
    import sys

    # first, just run it on the current module with default arguments:
    mod = sys.modules[init.__module__]
    _attach_self(mod)
    no_args = lambda mod: init(mod, *(), **{})  # type: ignore[call-arg]
    finalize(no_args, sys._getframe(1))

    # then attach it to the module's __dict__ for use when the module is mixed-in:
    init.__name__ = MIXIN_MODULE_INIT_HOOK_NAME
    init.__globals__[MIXIN_MODULE_INIT_HOOK_NAME] = init
    return init


def bases(mod: ModuleType | None = None) -> tuple[ModuleType, ...]:
    """
    Get the base modules for a module.

    If `mod` is None, determines the module from the caller's globals.
    This works with mixin's globals rewriting - the caller's globals()['__name__']
    will be the child module's name, so bases() returns the right bases.
    """
    if mod is None:
        import sys

        caller_globals = sys._getframe(1).f_globals
        mod_name = caller_globals['__name__']
        mod = sys.modules[mod_name]
    if b := getattr(mod, MIXIN_MODULE_BASES, None):
        return b
    return ()


def _add_base(mod: ModuleType, base: ModuleType) -> None:
    if bases := getattr(mod, MIXIN_MODULE_BASES, None):
        if base not in bases:
            bases = (*bases, base)
        updated_bases = tuple(bases)
        setattr(mod, MIXIN_MODULE_BASES, updated_bases)
    else:
        setattr(mod, MIXIN_MODULE_BASES, (base,))


def _rewrite_function_globals(
    func: FunctionType,
    new_globals: dict[str, Any],
    caller_mod_name: str,
    _seen: dict[int, FunctionType] | None = None,
) -> FunctionType:
    """
    create a copy of `func` with `new_globals`, recursively updating any
    functions captured in its closure.
    """
    if _seen is None:
        _seen = {}

    # avoid infinite recursion for self-referential closures
    if id(func) in _seen:
        return _seen[id(func)]

    # submodule gets priority lookup over the original module
    func_globals = chaindict(new_globals, func.__globals__)

    new_closure: tuple[CellType, ...] | None = None
    if func.__closure__:
        new_cells: list[CellType] = []
        for cell in func.__closure__:
            cell_contents = cell.cell_contents
            if isinstance(cell_contents, FunctionType):
                updated_func = _rewrite_function_globals(
                    cell_contents,
                    new_globals,
                    caller_mod_name,
                    _seen,
                )
                new_cells.append(CellType(updated_func))
            else:
                new_cells.append(cell)
        new_closure = tuple(new_cells)

    new_func = FunctionType(
        code=func.__code__,
        globals=func_globals,
        name=func.__name__,
        argdefs=func.__defaults__,
        closure=new_closure,
    )
    new_func.__dict__.update(func.__dict__)
    new_func.__annotations__ = dict(getattr(func, '__annotations__', {}))
    new_func.__module__ = caller_mod_name

    _seen[id(func)] = new_func
    return new_func


def _attach_self(mod: ModuleType | dict[str, Any]) -> None:
    if isinstance(mod, ModuleType):
        if hasattr(mod, MIXIN_MODULE_SELF):
            return
        setattr(mod, MIXIN_MODULE_SELF, mod)
    else:
        import sys

        if MIXIN_MODULE_SELF in mod:
            return
        mod[MIXIN_MODULE_SELF] = sys.modules[mod['__name__']]


type FinalizeCallback = Callable[[ModuleType], None]
_DEFERRED_FINALIZE_CALLBACKS: dict[FrameType, list[FinalizeCallback]] = defaultdict(list)


def finalize(transform: FinalizeCallback, caller_frame: FrameType | None = None):
    """run code after module is loaded"""
    if caller_frame is None:
        import sys

        caller_frame = sys._getframe(1)

    callbacks = _DEFERRED_FINALIZE_CALLBACKS[caller_frame]
    if transform in callbacks:
        return
    callbacks.append(transform)

    if caller_frame.f_trace is None:
        _attach_finalize_trace(caller_frame)


def _attach_finalize_trace(caller_frame: FrameType):
    import sys

    module_globals = caller_frame.f_globals
    _attach_self(module_globals)
    mod = module(module_globals)
    old_trace = sys.gettrace()

    def local_trace(frame, event, _arg):
        if frame is not caller_frame:
            return None
        if event in ('return', 'exception'):
            # cleanup: callbacks are popped, they have been executed and are no longer needed
            for callback in _DEFERRED_FINALIZE_CALLBACKS.pop(caller_frame, []):
                callback(mod)
            sys.settrace(old_trace)
            return None
        return local_trace

    def noop_trace(_frame, _event, _arg):
        return None

    caller_frame.f_trace = local_trace
    sys.settrace(noop_trace)



# fmt: off

from collections.abc import Callable, Iterable

__all__ = [
    'mklist',
    'mkset',
    'mkdict',
]

###############################################################################

def mklist[T](initial: Iterable[T] = ()) -> tuple[list[T], Callable[[T], None]]:
    """Returns a list and its bound append method."""
    lst = list(initial)
    return lst, lst.append

def mkset[T](initial: Iterable[T] = ()) -> tuple[set[T], Callable[[T], None]]:
    """Returns a dict and its bound add method."""
    st = set(initial)
    return st, st.add

def mkdict[K, V]() -> tuple[dict[K, V], Callable[[K, V], None]]:
    """Returns a dict and its bound setitem method."""
    dct = {}
    return dct, dct.__setitem__



from collections.abc import Callable
from types import FrameType, ModuleType, TracebackType

__all__ = [
    'frame_str',
    'frame_function',
    'frame_location',
    'frame_call_args',
    'exception_stack_strings',
    'exception_frame',
    'exception_location',
    'traceback_frame',
    'NO_FRAME_VALUE',
]


def frame_str(frame: FrameType, tag: bool = False) -> str:
    location = frame_location(frame)
    if not tag and location:
        return location
    if not location:
        return '<frame>'
    return f'<frame {location!r}>'


def frame_location(frame: FrameType | None, show_fun: bool = True) -> str | None:
    if not isinstance(frame, FrameType):
        return None

    line = frame.f_lineno
    code = frame.f_code
    file = code.co_filename
    name = code.co_qualname
    name = f'#{name}' if name and show_fun else ''
    file = f'{file}:{line}' if file and line else file
    return file + name


def frame_function(frame: FrameType | None) -> Callable | None:
    if not isinstance(frame, FrameType):
        return None

    qualname = frame.f_code.co_qualname
    if qualname == '<module>':
        return None
    if '.' in qualname:
        name_fst, name_rst = qualname.split('.', 1)
    else:
        name_fst = qualname
        name_rst = ''

    f_globals = frame.f_globals
    func = f_globals.get(name_fst, None)
    func = _getattr2(func, name_rst)
    if func is not None and callable(func):
        return func

    mod = frame_module(frame)
    if mod is None:
        return None

    func = _getattr2(mod, qualname)
    if callable(func):
        return func

    return None


NO_FRAME_VALUE = object()


def frame_call_args(frame: FrameType) -> list[tuple[str, object]]:
    from .code import code_arg_names

    get_local = frame.f_locals.get
    code = frame.f_code
    arg_names = code_arg_names(code)
    args = [(arg, get_local(arg, NO_FRAME_VALUE)) for arg in arg_names]
    if code.co_freevars:
        args.extend(_frame_closure(frame))
    return args


def _frame_closure(frame: FrameType) -> list[tuple[str, object]]:
    code = frame.f_code
    vars = code.co_freevars
    return [(f'#{var}', _frame_freevar(frame, var)) for var in vars]


def _frame_freevar(frame: FrameType, var: str) -> object:
    f: FrameType | None = frame
    while f:
        f_locals = f.f_locals
        if var in f_locals:
            return f_locals[var]
        f = f.f_back
    return NO_FRAME_VALUE


def frame_module(frame: FrameType) -> ModuleType | None:
    from sys import modules

    mod_name = frame.f_globals.get('__name__', '')
    if not isinstance(mod_name, str) or not mod_name:
        return None
    mod = modules.get(mod_name, None)
    if not isinstance(mod, ModuleType):
        return None
    return mod


DROPPED = '/threading.py:', '/concurrent/futures/thread.py:'


def exception_stack_strings(ex: BaseException) -> list[str]:
    stack = []
    frame = exception_frame(ex)
    while frame:
        loc = frame_location(frame)
        if not any(d in loc for d in DROPPED):
            stack.append(loc)
        frame = frame.f_back
    if len(stack) > 32:
        stack = stack[:16] + ['⋮'] + stack[-16:]
    return list(reversed(stack))


def exception_location(ex: BaseException) -> str | None:
    return frame_location(exception_frame(ex))


def exception_frame(ex: BaseException) -> FrameType | None:
    if not isinstance(ex, BaseException):
        return None
    return traceback_frame(ex.__traceback__)


def traceback_frame(tb: TracebackType | None) -> FrameType | None:
    if not isinstance(tb, TracebackType):
        return None
    while next := tb.tb_next:
        tb = next
    return tb.tb_frame


def _getattr2(obj: object, name: str) -> object:
    if obj is None:
        return None
    if not name:
        return obj
    if '.' not in name:
        return getattr(obj, name, None)
    for sub in name.split('.'):
        obj = getattr(obj, name, None)
        if obj is None:
            return None
    return obj



import uuid
from dataclasses import dataclass
from typing import Any, Literal, overload

from msgspec import Struct, field, to_builtins
from msgspec.json import Encoder

from ..core.datetime import time_now_utc

__all__ = [
    'CreateAgentRequest',
    'PromptTemplate',
    'should_template',
    'InteractionCodeBlock',
    'InteractionExecuteResult',
    'InteractionEvent',
    'ReasoningEffort',
    # sm messages
    'SMCreateAgentMessage',
    'SMDestroyAgentMessage',
    'SMInvocationEnterMessage',
    'SMInvocationExitMessage',
    'SMInvocationErrorMessage',
    'SessionManagerMessage',
    'SMInferenceRequestMessage',
    'SMInferenceResponseMessage',
    'SMInferenceErrorMessageMessage',
    'SMInvocationInteractionMessage',
    'SMMonadMessage',
    # monad messages
    'MonadStreamMessage',
    # magic messages
    'MagicMessage',
    'MagicExecuteEnterMessage',
    'MagicExecuteExitMessage',
    'MagicInferenceEnterMessage',
    'MagicInferenceExitMessage',
    'AllServerMessage',
    # encode
    'server_message_to_json',
    'server_message_to_dict',
]


# Request from the client

DEFAULT_PROTOCOL = 'python/v0'

# as per https://openrouter.ai/docs/guides/best-practices/reasoning-tokens#reasoning-effort-level
type ReasoningEffort = Literal['none', 'minimal', 'low', 'medium', 'high', 'xhigh']


class InteractionCodeBlock(Struct):
    code: str
    exec_id: str = field(default_factory=lambda: str(uuid.uuid4()))


class InteractionExecuteResult(Struct):
    result: str
    exec_id: str


type InteractionEvent = InteractionCodeBlock | InteractionExecuteResult


@dataclass(kw_only=True)
class PromptTemplate:
    template: str

    def __bool__(self) -> bool:
        return bool(self.template)


@overload
def should_template(value: PromptTemplate) -> str: ...
@overload
def should_template(value: str) -> None: ...
@overload
def should_template(value: None) -> None: ...
def should_template(value: str | PromptTemplate | None) -> str | None:
    if isinstance(value, str):
        return None
    if value is None:
        return None
    return value.template


@dataclass(kw_only=True)
class CreateAgentRequest:
    doc: str | None
    system: str | PromptTemplate | None
    model: str
    json: bool
    streaming: bool
    warp_globals_payload: bytes
    protocol: str = DEFAULT_PROTOCOL
    max_tokens_per_invocation: int | None = None
    max_tokens_per_round: int | None = None
    max_rounds: int | None = None
    reasoning_effort: ReasoningEffort | None = None


# Messages generated by the server


class SMCreateAgentMessage(Struct, tag="sm_create_agent"):
    uid: str
    doc: str | None
    system: str | None
    model: str
    json: bool
    streaming: bool
    timestamp: str = field(default_factory=time_now_utc)


class SMDestroyAgentMessage(Struct, tag="sm_destroy_agent"):
    uid: str
    timestamp: str = field(default_factory=time_now_utc)


class SMInvocationEnterMessage(Struct, tag="sm_invocation_enter"):
    uid: str
    iid: str
    timestamp: str = field(default_factory=time_now_utc)


class SMInvocationExitMessage(Struct, tag="sm_invocation_exit"):
    uid: str
    iid: str
    timestamp: str = field(default_factory=time_now_utc)


class SMInvocationErrorMessage(Struct, tag="sm_invocation_error"):
    uid: str
    iid: str
    error_type: str
    error_message: str | None = None
    timestamp: str = field(default_factory=time_now_utc)


class SMInferenceUsageMessage(Struct, tag="sm_inference_usage"):
    uid: str
    iid: str
    usage: dict[str, int]
    timestamp: str = field(default_factory=time_now_utc)


class SMMonadMessage(Struct, tag="sm_monad"):
    uid: str
    iid: str
    body: str
    timestamp: str = field(default_factory=time_now_utc)


class SMInvocationInteractionMessage(Struct, tag="sm_invocation_interaction"):
    uid: str
    iid: str
    event: InteractionEvent
    timestamp: str = field(default_factory=time_now_utc)


class SMInferenceRequestMessage(Struct, tag="sm_inference_request"):
    uid: str
    iid: str
    inference_id: str
    request: str
    timeout: int | None
    timestamp: str = field(default_factory=time_now_utc)


class SMInferenceResponseMessage(Struct, tag="sm_inference_response"):
    uid: str
    iid: str
    inference_id: str
    response: str
    timestamp: str = field(default_factory=time_now_utc)


class SMInferenceErrorMessageMessage(Struct, tag="sm_inference_error"):
    uid: str
    iid: str
    inference_id: str
    error_type: str
    error_message: str
    timestamp: str = field(default_factory=time_now_utc)


SessionManagerMessage = (
    SMInvocationEnterMessage
    | SMInvocationExitMessage
    | SMInvocationErrorMessage
    | SMInferenceErrorMessageMessage
    | SMInferenceRequestMessage
    | SMInferenceResponseMessage
    | SMInvocationInteractionMessage
)


# Messages generated by the agentic function invocation
class MagicExecuteEnterMessage(Struct, tag="mf_execute_enter"):
    uid: str
    iid: str
    code: str
    exec_id: str
    timestamp: str = field(default_factory=time_now_utc)


class MagicExecuteExitMessage(Struct, tag="mf_execute_exit"):
    uid: str
    iid: str
    exec_id: str
    out: str
    stdout: str
    stderr: str
    timestamp: str = field(default_factory=time_now_utc)


class MagicInferenceEnterMessage(Struct, tag="mf_inference_enter"):
    uid: str
    iid: str
    timestamp: str = field(default_factory=time_now_utc)


class MagicInferenceExitMessage(Struct, tag="mf_inference_exit"):
    uid: str
    iid: str
    response: str
    timestamp: str = field(default_factory=time_now_utc)


MagicMessage = (
    MagicExecuteEnterMessage
    | MagicExecuteExitMessage
    | MagicInferenceEnterMessage
    | MagicInferenceExitMessage
)


class MonadStreamMessage(Struct, tag="monad_stream"):
    uid: str
    iid: str
    content: str
    role: Literal['instruction', 'execution', 'code', 'agent']
    timestamp: str = field(default_factory=time_now_utc)


AllServerMessage = (
    SMMonadMessage
    | SMInvocationEnterMessage
    | SMInvocationExitMessage
    | SMInvocationErrorMessage
    | SMInferenceUsageMessage
    | SMInferenceRequestMessage
    | SMInferenceResponseMessage
    | SMInferenceErrorMessageMessage
    | SMInvocationInteractionMessage
    | MagicExecuteEnterMessage
    | MagicExecuteExitMessage
    | MagicInferenceEnterMessage
    | MagicInferenceExitMessage
    | SMCreateAgentMessage
    | SMDestroyAgentMessage
    | MonadStreamMessage
)


_enc = Encoder()


def server_message_to_json(message: AllServerMessage) -> str:
    return _enc.encode(message).decode("utf-8")


def server_message_to_dict(message: AllServerMessage) -> dict[str, Any]:
    # Use to_builtins for direct conversion instead of json.loads(json.dumps(...)) roundtrip
    return to_builtins(message)  # type: ignore[return-value]



from _collections import _tuplegetter  # type: ignore
from typing import ClassVar, Self, TypeGuard, overload

from ..hashing import *
from .code import AnsiCode

__all__ = [
    'Rgb',
    'is_rgb',
]


####################################################################################################


type Num = int | float


def is_rgb(obj: object) -> TypeGuard['Rgb']:
    return isinstance(obj, Rgb)


class Rgb(tuple[int, int, int]):
    r: int
    g: int
    b: int

    def __new__(cls, *args: int) -> Self:
        return tuple.__new__(cls, args)  # type: ignore

    def fmt_fg(self, o: object, w: int = 0) -> str:
        return self.fg(str(o).ljust(w))

    def fmt_hi(self, o: object, w: int = 0) -> str:
        return self.hi(str(o).ljust(w))

    __call__ = fmt_fg
    __matmul__ = fmt_hi

    def fg(self, s: str, d: float = 1.0) -> str:
        r, g, b = self.fg_tint(d)
        return f'\033[38;2;{r};{g};{b}m{s}\033[39m'

    def bg(self, s: str, d: float = 1.0) -> str:
        r, g, b = self.bg_tint(d)
        return f'\033[48;2;{r};{g};{b}m{s}\033[49m'

    def hi(self, s: str, f: float = 1.0, b: float = 1.0) -> str:
        (f_r, f_g, f_b), (b_r, b_g, b_b) = self.hi_tint(f, b)
        beg = f'\033[38;2;{f_r};{f_g};{f_b};48;2;{b_r};{b_g};{b_b}m'
        end = '\033[39;49m'
        return f'{beg}{s.replace(end, beg)}{end}'

    def darker(self, v: Num) -> Self:
        return self.lighter(-v)

    def lighter(self, v: Num) -> Self:
        r, g, b = self
        return Rgb(clip_i(r + v), clip_i(g + v), clip_i(b + v))

    def fg_tint(self, v: float = 1) -> Self:
        return self.bg_tint(-v)

    def bg_tint(self, v: float = 1) -> Self:
        r, g, b = self
        v *= Rgb.FG_DELTA if (r + g + b) > 120 else -Rgb.FG_DELTA
        return Rgb(clip_i(r - v), clip_i(g - v), clip_i(b - v))

    def hi_tint(self, fg: float = 1, bg: float = 1) -> tuple[Self, Self]:
        return self.lighter(10 * bg), self.to_lum(75 * fg)

    def __bool__(self) -> bool: ...

    @property
    def valid(self) -> bool:
        if not all(isinstance(arg, int) for arg in self):
            return False
        return min(self) >= 0 and max(self) <= 255

    def __repr__(self) -> str:
        if not self.valid:
            return f'Rgb({self.f_args()})'
        hex_str = repr(self.hex_str)
        if Rgb.COLOR_REPR:
            hex_str = self(hex_str)
        return f'Rgb.h({hex_str})'

    def __str__(self) -> str:
        if not self.valid:
            return self.f_args()
        hex_str = self.hex_str
        return self(hex_str) if Rgb.COLOR_REPR else hex_str

    def f_args(self) -> 'str':
        return ', '.join(str(arg) for arg in self)

    @overload
    def add(self, value: Self) -> Self: ...
    @overload
    def add(self, value: Num) -> Self: ...
    def add(self, value: Num | Self) -> Self:
        r1, g1, b1 = self
        if is_num(value):
            return Rgb(int(r1 + value), int(g1 + value), int(b1 + value))
        elif is_rgb(value):
            r2, g2, b2 = value
            return Rgb(r1 + r2, g1 + g2, b1 + b2)
        else:
            raise ValueError()

    def to_lum(self, l: Num) -> Self:
        return self.__mul__(l / self.lum_i)

    def mul(self, v: Num) -> Self:
        assert is_num(v)
        r, g, b = self
        return Rgb(clip_i(r * v), clip_i(g * v), clip_i(b * v))

    def div(self, v: Num) -> Self:
        return self.mul(1.0 / v)

    # these just let type inference know these methods exist
    __add__ = add
    __mul__ = mul
    __truediv__ = div

    def mix(self, other: Self, f: float = 0.5) -> Self:
        return Rgb.lerp(self, other, f)

    # constructors
    @staticmethod
    def iii(r: Num = 0, g: Num = 0, b: Num = 0) -> 'Rgb':
        assert is_nums(r, g, b), "invalid RGB ints {r}, {g}, {b}"
        return Rgb(clip_i(r), clip_i(g), clip_i(b))

    @staticmethod
    def fff(r: Num = 0, g: Num = 0, b: Num = 0) -> 'Rgb':
        assert is_nums(r, g, b), "invalid RGB floats {r}, {g}, {b})"
        return Rgb(clip_f(r), clip_f(g), clip_f(b))

    @staticmethod
    def i(i: int) -> 'Rgb':
        assert is_num(i), f"invalid grayscale int {i}"
        return _const(clip_i(i))

    @staticmethod
    def f(f: float) -> 'Rgb':
        assert is_num(f), f"invalid grayscale float {f}"
        return _const(clip_f(f))

    @staticmethod
    def h(s: str) -> 'Rgb':
        assert isinstance(s, str), f"invalid RGB hex {s}"
        assert len(s) == 6 and s.isalnum(), f"invalid RGB hex {s}"
        return Rgb(int(s[0:2], 16), int(s[2:4], 16), int(s[4:6], 16))

    @staticmethod
    def const(i: int) -> 'Rgb':
        return Rgb(i, i, i)

    @staticmethod
    def hex_list(spec: str) -> list['Rgb']:
        assert isinstance(spec, str), f"invalid hex list {spec})"
        return [Rgb.h(s) for s in spec.split()]

    @staticmethod
    def auto_int(i: int, f: float = 0.7) -> 'Rgb':
        if i == 0:
            return Rgb.GRAY
        if abs(i) < I10:
            return Rgb.uniq_int(i)
        if abs(i) < I48:
            return Rgb.id_int(i, f)
        return Rgb.hash_int(i, f)

    @staticmethod
    def uniq_int(i: int) -> 'Rgb':
        assert isinstance(i, int)
        if i == 0:
            return Rgb.GRAY
        elif i < 0:
            return Rgb.distinct(-i).darker(5)
        else:
            return Rgb.distinct(i - 1)

    @staticmethod
    def distinct(i: int) -> 'Rgb':
        cols = Rgb.DISTINCT
        return cols[abs(i) % len(cols)]

    @staticmethod
    def hash_int(h: int, f: float = 0.7) -> 'Rgb':
        assert isinstance(h, int) and isinstance(f, float)
        r, g, b = (abs(h) & 0xFFFFFF).to_bytes(3)
        return Rgb(hash_i(r, f), hash_i(g, f), hash_i(b, f))

    @staticmethod
    def hash_bytes(b: bytes, f: float = 0.7) -> 'Rgb':
        return Rgb.hash_int(raw_bytes_hash(b), f)

    @staticmethod
    def hash_str(s: str, f: float = 0.7) -> 'Rgb':
        return Rgb.hash_int(raw_str_hash(s), f)

    @staticmethod
    def id_int(h: int, f: float = 0.7) -> 'Rgb':
        assert isinstance(h, int)
        return Rgb.hash_int(raw_int_hash(h), f)

    @staticmethod
    def id_obj(obj: object, f: float = 1.0) -> 'Rgb':
        return Rgb.hash_int(raw_int_hash(id(obj)), f)

    @staticmethod
    def hex(s: str) -> 'Rgb':
        s = s.lstrip('#')
        if len(s) == 6:
            return Rgb.h(s)
        elif len(s) == 3:
            return Rgb.h(s[0] * 2 + s[1] * 2 + s[2] * 2)
        else:
            raise ValueError(f"invalid hex color {s}")

    @staticmethod
    def mean(*cols: 'Rgb') -> 'Rgb':
        n = len(cols)
        if n == 0:
            return Rgb.GRAY
        r = sum(rgb[0] for rgb in cols) / n
        g = sum(rgb[1] for rgb in cols) / n
        b = sum(rgb[2] for rgb in cols) / n
        return Rgb(clip_i(r), clip_i(g), clip_i(b))

    @staticmethod
    def lerp(c1: 'Rgb', c2: 'Rgb', f1: float) -> 'Rgb':
        r1, g1, b1 = c1
        r2, g2, b2 = c2
        f2 = 1.0 - f1
        r = mix_i(r1, r2, f1, f2)
        g = mix_i(g1, g2, f1, f2)
        b = mix_i(b1, b2, f1, f2)
        return Rgb(r, g, b)

    @property
    def ansi_f(self) -> AnsiCode:
        r, g, b = self
        return AnsiCode(f'38;2;{r};{g};{b}')  # type: ignore

    @property
    def ansi_b(self) -> AnsiCode:
        r, g, b = self
        return AnsiCode(f'48;2;{r};{g};{b}')  # type: ignore

    @property
    def lum_f(self) -> float:
        r, g, b = self
        return r * 0.212 + g * 0.715 + b * 0.0722

    @property
    def lum_i(self) -> float:
        return clip_i(self.lum_f * 255)

    @property
    def is_white(self) -> bool:
        return min(self) >= 255

    @property
    def is_black(self) -> bool:
        return min(self) <= 0

    @property
    def is_nonzero(self) -> bool:
        return bool(max(self)) or bool(min(self))

    @property
    def hex_str(self) -> str:
        r, g, b = self
        return f"#{r:02X}{g:02X}{b:02X}"

    @property
    def floats(self) -> tuple[float, float, float]:
        r, g, b = self
        return r / 255.0, g / 255.0, b / 255.0

    # fmt: off
    GRAY:       ClassVar['Rgb']
    WHITE:      ClassVar['Rgb']
    BLACK:      ClassVar['Rgb']
    TRUE:       ClassVar['Rgb']
    FALSE:      ClassVar['Rgb']

    DISTINCT:   ClassVar[list['Rgb']]

    FG_DELTA:   int = 80
    COLOR_REPR: bool = True
    # fmt: on


####################################################################################################


def _const(i: int) -> Rgb:
    return Rgb(i, i, i)


def _rgb(r: int, g: int, b: int) -> Rgb:
    return Rgb(clip_i(r), clip_i(g), clip_i(b))


# setup r, g, b accessors

Rgb.r = property(_tuplegetter(0, 'r'))  # type: ignore
Rgb.g = property(_tuplegetter(0, 'g'))  # type: ignore
Rgb.b = property(_tuplegetter(0, 'b'))  # type: ignore

Rgb.__bool__ = Rgb.is_nonzero  # type: ignore

# utilities


def mix_i(n1: Num, n2: Num, f1: Num, f2: Num) -> int:
    return clip_i(f1 * n1 + f2 * n2)


def clip_i(i: int | float) -> int:
    return min(max(round(i), 0), 255)


def clip_f(i: int | float) -> int:
    return min(max(round(i * 255), 0), 255)


def hash_i(v: int, f: float):
    return round(v * f + 127 * (1 - f))


def is_num(arg: object) -> TypeGuard[Num]:
    return type(arg) in (int, float)


def is_nums(*args: object) -> bool:
    return all(type(a) in (int, float) for a in args)


####################################################################################################

# setup palettes

# fmt: off
Rgb.BLACK    = Rgb(0, 0, 0)
Rgb.GRAY     = Rgb(128, 128, 128)
Rgb.WHITE    = Rgb(255, 255, 255)
Rgb.TRUE     = Rgb.i(245)
Rgb.FALSE    = Rgb.i(10)
Rgb.DISTINCT = Rgb.hex_list('1f77b4 ff7f0e 2ca02c d62728 9467bd 8c564b e377c2 7f7f7f bcbd22 17becf aec7e8 ffbb78 98df8a ff9896 c5b0d5 c49c94 f7b6d2 c7c7c7 dbdb8d 9edae5')
# fmt: on

I10 = 1 << 10
I24 = 1 << 24
I48 = 1 << 48
# fmt: off



# fmt: off

import re

from ..repl import ReplP

from .__ import *
from .base import Request

__all__ = [
    'ReplRequest',
    'ReplInit',
    'ReplRunCode',
    'ReplCallMethod',
]


################################################################################

if TYPE_CHECKING:
    from ...repl.repl_eval_info import ReplEvaluationInfo

################################################################################


class ReplRequest(Request, ABC):
    """
    ABC for a fully decoded REPL request.
    """

    __slots__ = __match_args__ = ('repl',)

    repl: 'ReplP'

    ############################################################################

    def __fmt_args__(self):
        for slot in self.__slots__[1:]:
            yield f_object_id(getattr(self, slot, FIELD_ABSENT))

    ############################################################################

    def execute(self) -> Result:
        log = bool(LOG_REPL)
        if not hasattr(self, "repl"):
            P.nprint(ICON_E, 'no repl present') if log else None
            return Result.bad(RuntimeError("no repl present"))
        try:
            val = self.__execute__(log)
            res = Result.good(val)
        except E.WarpError as err:
            P.nprint(ICON_E, err, sep='\n') if log else None
            res = Result.bad(RuntimeError(f"internal error: {err}"))
        except BaseException as err:
            P.nprint(ICON_E, err, sep='') if log else None
            res = Result.bad(err)
        P.nprint(ICON_O, res, '\n') if log else None
        return res

    @abstractmethod
    def __execute__(self, log: bool) -> TermT: ...

    ############################################################################

    async def execute_async(self) -> Result:
        log = bool(LOG_REPL)
        if not hasattr(self, "repl"):
            P.nprint(ICON_E, 'no repl present') if log else None
            return Result.bad(RuntimeError("no repl present"))
        try:
            val = await self.__execute_async__(log)
            res = Result.good(val)
        except E.WarpError as err:
            P.nprint(ICON_E, err, sep='\n') if log else None
            res = Result.bad(RuntimeError(f"internal error: {err}"))
        except BaseException as err:
            P.nprint(ICON_E, err, sep='') if log else None
            res = Result.bad(err)
        P.nprint(ICON_O, res, '\n') if log else None
        return res

    async def __execute_async__(self, log: bool) -> TermT:
        return self.__execute__(log)

    ############################################################################

    @abstractmethod
    def encode(self, codec: 'EncoderP') -> 'ReplRequestMsg': ...


################################################################################

class ReplInit(ReplRequest):

    __slots__ = __match_args__ = 'repl', 'globals', 'locals'

    globals: Rec[TermT]
    locals: Rec[TermT]

    def __init__(self, globals_: Rec[TermT], locals_: Rec[TermT]):
        self.globals = globals_
        self.locals = locals_

    def __execute__(self, log: bool):
        if log:
            P.nprint(ICON_I, f'ReplInit({len(self.globals)}.., {len(self.locals)}..)')

        ORDER_MARKER = '|'
        hidden_set: set[str] = set()

        def process_vars(dct: dict[str, object]) -> dict[str, object]:

            if '__hidden_names' in dct:
                hidden = dct.get('__hidden_names')
                if isinstance(hidden, (list, tuple, set)) and all(isinstance(x, str) for x in hidden):
                    hidden_set.update(hidden)
                    dct = {k: v for k, v in dct.items() if k != '__hidden_names'}

            if dct and any(ORDER_MARKER in k for k in dct.keys()):
                key_map: list[tuple[int, str, str]] = []
                for marked_name in dct.keys():
                    if ORDER_MARKER in marked_name:
                        name, idx = marked_name.split(ORDER_MARKER, 1)
                        try:
                            pos = int(idx)
                        except Exception:
                            pos = 0
                        key_map.append((pos, name, marked_name))
                    else:
                        key_map.append((10**9, marked_name, marked_name))

                ordered: dict[str, object] = {}
                for _, name, orig in sorted(key_map, key=lambda t: t[0]):
                    ordered[name] = dct[orig]
                return ordered

            return dct

        global_vars = process_vars(self.globals or {})
        local_vars = process_vars(self.locals or {})

        if log:
            for k, v in global_vars.items():
                P.nprint(ICON_M, f'will set global {k!r} =', f_object_id(v))

            for k, v in local_vars.items():
                P.nprint(ICON_M, f'will set local {k!r} =', f_object_id(v))

        return self.repl.initialize(
            local_vars=local_vars,
            global_vars=global_vars,
            hidden_vars=tuple(hidden_set),
        )

    def encode(self, enc):
        from ..msg.rpc_request_repl import ReplInitMsg
        return ReplInitMsg(
            enc.enc_raw(self.globals),
            enc.enc_raw(self.locals),
        )


################################################################################

# false positives are fine!

async_hints = (
    'await', 'async', 'asyncio', 'run', 'run_until_complete', 'gather',
    'get_event_loop', 'new_event_loop',
    'current_task', 'create_task', 'as_completed', 'wait', 'wait_for',
    'Future', 'Task', 'TaskGroup'
)

ASYNC_SOURCE_RE = re.compile(rf'\b({'|'.join(async_hints)})\b')

class ReplRunCode(ReplRequest):

    __slots__ = __match_args__ = 'repl', 'source', 'options',

    source:  str
    options: KwargsT

    def __init__(self, source: str, **options):
        self.source = source
        self.options = options

    @property
    def is_async(self) -> bool:
        # This used to check for do:
        #return bool(ASYNC_SOURCE_RE.match(self.source))
        # since false positives are completely fine.
        # For now, to avoid having to deal with different logic for each, simply return True.
        #return 'await ' in self.source or 'async ' in self.source
        return True

    def __execute__(self, log: bool) -> 'ReplEvaluationInfo':
        self.__log_input__(log)
        return self.repl.run_code_info(self.source, **self.options)

    async def __execute_async__(self, log: bool) -> TermT:
        self.__log_input__(log)
        info = await self.repl.async_run_code_info(self.source, **self.options)
        return info

    def __log_input__(self, log: bool):
        if not log:
            return
        f_source = self.source.strip('\n')
        f_source = FMT_CODE(f_source)
        if '\n' in f_source:
            P.nprint(ICON_I, f"'''\n{f_source}\n'''")
        else:
            P.nprint(ICON_I, f_source)

    def encode(self, enc):
        from ..msg.rpc_request_repl import ReplRunCodeMsg
        return ReplRunCodeMsg(self.source, self.options)


################################################################################

class ReplCallMethod(ReplRequest):

    __slots__ = __match_args__ = 'repl', 'method', 'pos', 'key'

    method: str
    pos:    ArgsT
    key:    KwargsT

    def __init__(self, method: str, pos: ArgsT, key: KwargsT):
        self.method = method
        self.pos = pos
        self.key = key

    def __execute__(self, log: bool):
        if log:
            P.nprint(ICON_I, f'repl.{self.method}(', *self.pos, ')')
        fn = getattr(self.repl, self.method)
        return fn(*self.pos, **self.key)

    def encode(self, enc):
        from ..msg.rpc_request_repl import ReplCallMethodMsg
        return ReplCallMethodMsg(self.method, enc.enc_args(self.pos), enc.enc_kwargs(self.key))


################################################################################

LOG_REPL = LogFlag('REPL')

color = P.MEDIUM.R
ICON_I = color @ '<<<'
ICON_Q = color @ '`'
ICON_M = color @ '---'
ICON_O = color @ '>>>'
ICON_E = color @ '!!!'
FMT_CODE = color



from __future__ import annotations

import asyncio as aio

from agentica_internal.core.print import tprint
from agentica_internal.warpc.messages import DefinitionMsg, RPCMsg
from agentica_internal.warpc.repl import DummyRepl
from agentica_internal.warpc.worlds.sdk_world import SDKWorld
from agentica_internal.warpc_transcode.playground.toy_translator import Translator


async def main() -> None:
    translator = Translator()

    async def TS_recv_bytes() -> bytes:
        msg: RPCMsg = await aio.to_thread(translator.recv_py_rpc, 1000)
        tprint("Really received Python RPCMsg:", msg)
        return msg.to_msgpack()

    async def TS_send_bytes(data: bytes) -> None:
        tprint("TS_send_bytes", data)
        msg = RPCMsg.from_msgpack(data)
        tprint("Preparing to send Python RPCMsg:", msg)
        await aio.to_thread(translator.send_py_rpc, msg)

    sandbox = SDKWorld(logging=True)
    sandbox.attach_repl(DummyRepl())

    # Init REPL with meaning
    if translator.init:
        print("Initializing REPL with", translator.init)
        decoded: dict[str, object] = {}
        for name, res_def in translator.init.items():
            if isinstance(res_def, DefinitionMsg):
                decoded[name] = sandbox.frames.root.dec_any(res_def)
        repl = DummyRepl()
        sandbox.attach_repl(repl)
        repl.set_global_vars(decoded)
        print("REPL Initialized with", decoded)

    print("Sandbox after init:", sandbox)

    # Call meaning
    assert sandbox.frames.repl is not None
    meaning = sandbox.frames.repl.get_var("meaningOfLife")
    sandbox.event_loop = aio.get_running_loop()
    async with aio.TaskGroup() as tg:
        tg.create_task(sandbox.drain_outbox(TS_send_bytes))
        tg.create_task(sandbox.fill_inbox(TS_recv_bytes))
        tg.create_task(sandbox.process_inbox(tg))

        def _call_meaning():
            with sandbox.frames.root:
                return meaning()  # type: ignore

        result = await aio.to_thread(_call_meaning)
        tprint("Result:", result)


if __name__ == "__main__":
    aio.run(main())



# fmt: off

from .__ import *

from .handle import ResourceHandle, has_handle, get_handle
from .virtual_class import ClassData
from .virtual_coroutine import CoroutineData
from .virtual_function import FunctionData
from .virtual_iterator import IteratorData
from .virtual_module import ModuleData
from .virtual_object import ObjectData
from .virtual_type import TypeData, CallableTypeData, TypeVarData, TypeAliasData, GenericAliasData, TypeUnionData, ForwardRefData
from .virtual_async import FutureData, unregister_virtual_future
from .virtual_enum import EnumKind, EnumClassData
from .base import ResourceData

__all__ = [
    'ResourceHandle',
    'has_handle',
    'get_handle',
    'ResourceData',
    'ObjectData',
    'ClassData',
    'FunctionData',
    'CoroutineData',
    'ModuleData',
    'TypeData',
    'CallableTypeData',
    'TypeVarData',
    'TypeAliasData',
    'GenericAliasData',
    'TypeUnionData',
    'ForwardRefData',
    'IteratorData',
    'FutureData',
    'EnumClassData',
    'EnumKind',
    'unregister_virtual_future',
]



# EXCLUDE_FROM_SDK_BUILD
# fmt: off

from ..__ import *

from copy import copy
from pathlib import Path

from agentica_internal.cpython.alias import *
from agentica_internal.cpython.shed.load import get_shed_class, get_shed_module
from agentica_internal.warpc.attrs import *

from agentica_internal.warpc.resource.virtual_function import make_template_strs, describe_real_function

type OClass = type
type SClass = type
type VClass = type

type Kind = Literal['i', 'c', 's']
type Methods = dict[str, tuple[FunctionT, Kind]]


def is_func(func: Any) -> TypeGuard[FunctionT]:
    return type(func) is FunctionType


SELF_REF = f"get_raw(self, '{VHDL}')"
TYPE_REF = f"get_raw(type(self), '{VHDL}')"
CLS_REF = f"get_raw(self, '{VHDL}')"


class MethodInfo:
    name: str
    func: FunctionType
    kind: Kind
    code: str

    DROP: str = ''

    def __init__(self, func: Callable, code: str):
        assert is_func(func)
        self.name = func.__name__
        self.func = func
        self.code = code

    def __repr__(self) -> str:
        f_cls = type(self).__name__
        return f'{f_cls}({self.name!r}, {self.code!r})'

    @staticmethod
    def from_shed(func: Callable) -> 'MethodInfo':
        cls = type(func)
        func = getattr(func, '__wrapped__', func)
        code = getattr(func, '___shed_src___')
        info = TO_METHOD_INFO_CLS[cls](func, code)
        info.fixup()
        return info

    def to_virtual(self, cls_name: str) -> 'MethodInfo':
        v_info = copy(self)
        fn_def = describe_real_function(self.func)
        if type(self) is not SMethodInfo:
            if fn_def.pos_args:
                fn_def.pos_args = fn_def.pos_args[1:]
            elif fn_def.args:
                fn_def.args = fn_def.args[1:]
        _, pos_str, key_str = make_template_strs(fn_def)
        args_str = f'{pos_str}, {key_str}'
        args = pos_str[1:-1].split(', ')
        v_callback_str = self.v_callback_str(cls_name, args_str, *args)
        v_info.code = replace_ellipsis(self.code, v_callback_str)
        return v_info

    def v_callback_str(self, cls_name: str, args_str: str, *args: str) -> str: ...

    def fixup(self):
        pass


SPECIAL = (
    '__setattr__',
    '__getattr__',
    '__delattr__',
)

SYS_METHODS = {
    '__str__': 'str',
    '__len__': 'len',
    '__iter__': 'iter',
    '__aiter__': 'aiter',
    '__next__': 'next',
    '__anext__': 'anext',
    '__repr__': 'str',
    '__hash__': 'hash',
    '__copy__': 'copy',
}


class IMethodInfo(MethodInfo):
    def v_callback_str(self, cls_name: str, args_str: str, *args) -> str:
        return f"return v_callback('callmethod', {SELF_REF}, '{self.name}', {args_str})"

    def fixup(self):
        if self.name in SPECIAL:
            self.__class__ = CustomMethodInfo
        if sys_meth := SYS_METHODS.get(self.name):
            self.__class__ = SysMethodInfo
            self.sys_meth = sys_meth


class CMethodInfo(MethodInfo):
    def v_callback_str(self, cls_name: str, args_str: str, *args) -> str:
        return f"return v_callback('callmethod', {CLS_REF}, {args_str})"


class SMethodInfo(MethodInfo):
    def v_callback_str(self, cls_name: str, args_str: str, *args) -> str:
        return 'raise NotImplementedError()'

    def fixup(self):
        if self.name == '__new__':
            self.__class__ = CustomMethodInfo


class CustomMethodInfo(MethodInfo):
    def v_callback_str(self, cls_name: str, args_str: str, *args: str) -> str:
        match self.name:
            case '__new__':
                return f"return v_callback('new', {CLS_REF}, {args_str})"
            case '__setattr__':
                return f"return v_callback('setattr', {SELF_REF}, {args[0]}, {args[1]})"
            case '__getattr__':
                return f"return v_callback('getattr', {SELF_REF}, {args[0]})"
            case '__delattr__':
                return f"return v_callback('delattr', {SELF_REF}, {args[0]})"
        raise ValueError(self)


class SysMethodInfo(MethodInfo):
    sys_meth: str

    def v_callback_str(self, cls_name: str, args_str: str, *strs) -> str:
        return f"return v_callback('callsysmethod', {SELF_REF}, {self.sys_meth})"


TO_METHOD_INFO_CLS = {
    FunctionT: IMethodInfo,
    classmethod: CMethodInfo,
    staticmethod: SMethodInfo,
}


def replace_ellipsis(code: str, body: str) -> str:
    *most, last = code.rstrip(' \n\t').rsplit('\n', 1)
    last = last.removesuffix('...')
    margin = last.split('def', 1)[0]
    most = '\n'.join(most)
    return f'{most}\n{last}\n{margin}    {body}'


class ClassInfo:
    cls: OClass
    name: str
    module: str | None
    rename: str | None
    bases: tuple[str, ...]
    meths: dict[str, MethodInfo]
    slots: tuple[str, ...] | None

    def __init__(self, cls: OClass, name: str, module: str | None):
        self.cls = cls
        self.name = name
        self.module = module
        self.meths = {}
        self.bases = ()
        self.module = module
        self.rename = None
        self.slots = None

    def __repr__(self) -> str:
        return f'ClassInfo({self.name!r}, {self.bases!r}, {self.meths!r})'

    @staticmethod
    def from_shed(s_cls: SClass) -> 'ClassInfo':
        o_cls = s_cls.___shed_ori___
        info = ClassInfo(o_cls, o_cls.__name__, o_cls.__module__)
        info.bases = tuple(b.__name__ for b in o_cls.__bases__)
        info.slots = getattr(o_cls, '__slots__', None)
        add = info.meths.__setitem__
        for key, val in s_cls.__dict__.items():
            if key in IGNORE:
                continue
            if isinstance(val, (FunctionType, classmethod, staticmethod)):
                m_info = MethodInfo.from_shed(val)
                add(key, m_info)
        return info

    def update(self, other: 'ClassInfo') -> None:
        self.meths.update(other.meths)

    def to_virtual(self) -> Self:
        s_name = self.name
        v_name = 'virtual_' + s_name
        v_info = ClassInfo(self.cls, v_name, self.module)
        v_info.bases = (s_name,)
        v_info.meths = {k: v.to_virtual(s_name) for k, v in self.meths.items()}
        v_info.rename = s_name
        return v_info

    def code(self) -> str:
        return '\n\n'.join(self.codelines())

    def codelines(self):
        bases = ', '.join(self.bases)
        name = self.name
        yield f'class {self.name}({bases}):'
        if slots := self.slots:
            yield f'    __slots__ = {slots!r}'
        for meth in self.meths.values():
            yield meth.code
        if rename := self.rename:
            yield f'{name}.__name__ = {name}.__qualname__ = {rename!r}'
        if module := self.module:
            yield f'{name}.__module__ = {module!r}'


class SClassInfo(dict[OClass, ClassInfo]):
    def __missing__(self, key: OClass) -> ClassInfo:
        return self.add_class(key)

    def add_class(self, cls: OClass) -> ClassInfo:
        s_cls = get_shed_class(cls)
        s_info = ClassInfo.from_shed(s_cls)
        self[cls] = s_info
        for base in s_cls.__bases__:
            b_info = SCLASS_INFO[base]
            s_info.update(b_info)
        return s_info


class VClassInfo(dict[SClass, ClassInfo]):
    def __missing__(self, key: OClass) -> ClassInfo:
        return self.add_class(key)

    def add_class(self, cls: OClass) -> ClassInfo:
        s_info = SCLASS_INFO[cls]
        v_info = s_info.to_virtual()
        self[cls] = v_info
        return v_info


IGNORE = (
    '__init_subclass__',
    '__init__',
    '__getattribute__',
    '__reduce__',
    '__reduce_ex__',
    '__getstate__',
    # type
    '__prepare__',
    '__subclasscheck__',
    '__subclasshook__',
    '__sizeof__',
    '__subclasses__',
    'mro',
)

SCLASS_INFO = SClassInfo()
VCLASS_INFO = VClassInfo()


def generate():
    this_dir = Path(__file__).parent
    builtins_mod = get_shed_module('builtins')

    out_file = this_dir / 'generated.py'

    v_blocks, add_block = mklist()
    add_block('from .common import *')

    for cls in [object, type, list, set]:
        v_info = VCLASS_INFO[cls]
        v_code = v_info.code()

        P.hprint(v_code)
        add_block(v_code)

    out_file.write_text('\n\n'.join(v_blocks))


generate()



# fmt: off

from enum import IntEnum

from asyncio import Future
from types import NoneType
from typing import Protocol, _ProtocolMeta, Any

from ...javascript import Number
from .term_resource import class_to_system_msg, pointer_to_system_msg


__all__ = [
    'OBJECT',
    'TYPE',
    'STR',
    'BOOL',
    'INT',
    'LIST',
    'DICT',
    'SET',
    'TUPLE',
    'FUTURE',
    'BASE_EXCEPTION',
    'EXCEPTION',
    'PROTOCOL',
    'PROTOCOL_META',
    'NONE_TYPE',
    'ANY',
    'NUMBER',
    'IDs',
]


################################################################################

OBJECT         = class_to_system_msg(object)
TYPE           = class_to_system_msg(type)
STR            = class_to_system_msg(str)
BOOL           = class_to_system_msg(bool)
INT            = class_to_system_msg(int)
LIST           = class_to_system_msg(list)
DICT           = class_to_system_msg(dict)
SET            = class_to_system_msg(set)
TUPLE          = class_to_system_msg(tuple)

FUTURE         = class_to_system_msg(Future)
BASE_EXCEPTION = class_to_system_msg(BaseException)
EXCEPTION      = class_to_system_msg(Exception)
PROTOCOL       = class_to_system_msg(Protocol)
PROTOCOL_META  = class_to_system_msg(_ProtocolMeta)
NONE_TYPE      = class_to_system_msg(NoneType)

ANY            = pointer_to_system_msg(id(Any))
NUMBER         = pointer_to_system_msg(id(Number))

################################################################################

class IDs(IntEnum):
    Type            = TYPE.sid
    Object          = OBJECT.sid
    Str             = STR.sid
    Bool            = BOOL.sid
    Int             = INT.sid
    List            = LIST.sid
    Dict            = DICT.sid
    Set             = SET.sid
    Tuple           = TUPLE.sid
    BaseException   = BASE_EXCEPTION.sid
    Exception       = EXCEPTION.sid
    Protocol        = PROTOCOL.sid
    NoneType        = NONE_TYPE.sid
    Any             = ANY.sid



import pytest
from agentica_internal.testing.warpc.forbidden_tests import *


@pytest.mark.asyncio
async def test_forbidden_module():
    await verify_forbidden_module()


@pytest.mark.asyncio
async def test_forbidden_whitelist():
    await verify_forbidden_whitelist()



# fmt: off

from .annos import ANNOS
from .classes import CLASSES
from .frames import FRAMES
from .funcs import FUNCTIONS
from .methods import METHODS
from .modules import MODULES
from .objects import OBJECTS

__ALL__ = [
    'VALUES',
]

VALUES = ANNOS + CLASSES + FUNCTIONS + METHODS + MODULES + FRAMES + OBJECTS



import atexit

__all__ = [
    'ex_function',
    'ex_generator_fn',
    'ex_generator',
    'ex_async_generator',
    'ex_async_generator_fn',
    'ex_coroutine_fn',
    'ex_coroutine',
    'ex_code',
    'ex_cell',
    'ex_object',
    'ex_slot_object',
    'ex_static_method',
    'ex_bound_class_method',
    'ex_mapping_proxy',
    'ex_property',
    'ex_slot_property',
    'ex_unbound_class_method',
    'ex_unbound_method',
    'ex_bound_method',
    'ex_bound_super_method',
    'ex_exception',
    'ex_traceback',
    'ex_frame',
    'ex_type_alias',
    'ex_union',
    'ex_generic_alias',
    'ex_builtin_function',
    'ex_class_method_descriptor',
    'ex_get_set_descriptor',
    'ex_method_descriptor',
    'ex_method_wrapper',
    'ex_singleton',
    'ex_wrapper_descriptor',
]  # noqa

################################################################################


def ex_function(x: int, y: bool = True, *args, **kwargs) -> float:
    """ex_function docstring"""
    return 1


def ex_generator_fn():
    """ex_generator_fn docstring"""
    yield 1


async def ex_async_generator_fn():
    """ex_async_generator_fn docstring"""
    yield 1


async def ex_coroutine_fn():
    """ex_coroutine_fn docstring"""
    pass


################################################################################

ex_async_generator = ex_async_generator_fn()
ex_generator = ex_generator_fn()
ex_coroutine = ex_coroutine_fn()

atexit.register(lambda: ex_coroutine.close())

################################################################################

ex_code = ex_function.__code__


def ex_outer_capturing_fn():
    """ex_outer_capturing_fn docstring"""
    value = 1

    def ex_inner_capturing_fn():
        """ex_inner_capturing_fn docstring"""
        nonlocal value

    return ex_inner_capturing_fn.__closure__[0]  # type: ignore


ex_cell = ex_outer_capturing_fn()

################################################################################


class ExClass:
    """ExClass docstring"""

    x: int

    def __init__(self):
        self.x = 1
        self.y = True

    @property
    def ex_instance_prop(self) -> int:
        """ExClass.ex_instance_prop docstring"""
        return self.x

    def ex_instance_method(self, arg1: str) -> int:
        """ExClass.ex_instance_method docstring"""
        return self.x + len(arg1) + 1

    @staticmethod
    def ex_static_method(arg1: str) -> int:
        """ExClass.ex_static_method docstring"""
        return len(arg1) + 1

    @classmethod
    def ex_class_method(cls, arg1: str) -> int:
        """ExClass.ex_class_method docstring"""
        return len(cls.__name__) + len(arg1) + 1


class ExSlotClass:
    """ExSlotClass docstring"""

    __slots__ = (
        'slot1',
        'slot2',
    )


class ExException(Exception):
    """ExException docstring"""

    pass


################################################################################

ex_object = ExClass()
ex_slot_object = ExSlotClass()
ex_static_method = ExClass.ex_static_method
ex_bound_class_method = ExClass.ex_class_method
ex_mapping_proxy = ExClass.__dict__
ex_property = ExClass.ex_instance_prop
ex_slot_property = ExSlotClass.slot1  # type: ignore
ex_unbound_class_method = ex_mapping_proxy['ex_class_method']
ex_unbound_method = ex_mapping_proxy['ex_instance_method']
ex_bound_method = ex_object.ex_instance_method
ex_bound_super_method = ex_object.__str__

################################################################################

try:
    raise ExException(1)
except ExException as exc:
    ex_exception = exc
    ex_traceback = exc.__traceback__
    ex_frame = exc.__traceback__.tb_frame  # type: ignore

################################################################################

type ex_type_alias = ExClass | str

ex_union = ExClass | str

ex_generic_alias = list[ExClass]

################################################################################

# builtin function where `__self__ = builtins`
ex_builtin_function = len

# builtin function where `__self__ = int`
ex_builtin_method = int.__new__

ex_class_method_descriptor = bytes.__dict__['fromhex']

ex_get_set_descriptor = type(ex_function).__code__  # type: ignore

# bound class method of an CPython object
ex_method_descriptor = bytes.hex

# unbound instance method of an CPython object
ex_wrapper_descriptor = bytes.__buffer__

# bound instance method of an CPython object
ex_method_wrapper = bytes().__buffer__

ex_member_descriptor = range.start

#######################################################ex#########################

ex_singleton = NotImplemented



# fmt: off

################################################################################

from .__json import pprint_json, fmt_json
from .__msgpack import pprint_msgpack, fmt_msgpack

################################################################################

from .base import Msg

from .bad import BadMsg, NO_MSG, CORRUPT_MSG, UNEXPECTED_MSG, is_bad_msg

from .codec import (
    EncoderP,
    EncoderContextP,
    DecoderP,
    DecoderContextP,
    CodecP,
    NULL_CODEC
)

################################################################################

from .term import (
    TermMsg,
    TermPassByRefMsg,
    TermPassByValMsg,
)

from .term_exception import ExceptionMsg

from .term_atom import (
    AtomMsg,
    NumberMsg, SingletonMsg, NoneMsg,
    StrLikeMsg, StrMsg, BytesMsg,
    NotImplMsg, EllipsisMsg, ClosedMsg
)

from .term_container import (
    ContainerMsg,
    MappingMsg, DictMsg,
    SequenceMsg, ListMsg, TupleMsg, SetMsg, FrozenSetMsg,
)

from .term_special import (
    SlotObjMsg,
    ReduceObjMsg,
    RegexPatternMsg,
    RegexMatchMsg,
    ClassUnionMsg,
    EnumMemberMsg,
    EnumKeyMsg,
    EnumValMsg
)

from .term_resource import (
    ResourceMsg,
    UserResourceMsg,
    SystemResourceMsg,
    RemoteResourceMsg,
    LocalResourceMsg,
    class_to_system_msg,
    class_to_system_id,
    pointer_to_system_id,
    pointer_to_system_msg,
)


from .term_lambda import LambdaMsg, SyntaxMsg

################################################################################

from .resource_def import DefinitionMsg

from .resource_data import (
    ResourceDataMsg,
    ObjectDataMsg, ClassDataMsg, FunctionDataMsg, ModuleDataMsg, CoroutineDataMsg,
    TypeDataMsg, TypeAliasDataMsg, TypeVarDataMsg, GenericAliasDataMsg,
    CallableTypeDataMsg, TypeUnionDataMsg, IteratorDataMsg, EnumClassDataMsg
)

################################################################################

from .rpc import RPCMsg

from .rpc_framed import FramedRequestMsg, FramedResponseMsg

################################################################################

from .rpc_request import RequestMsg

from . import rpc_legacy as legacy

from .rpc_request_repl import (
    ReplRequestMsg,
    ReplInitMsg,
    ReplCallMethodMsg,
    ReplRunCodeMsg,
)

from .rpc_request_resource import (
    ResourceRequestMsg,
    ResourceNewMsg,
    ResourceCallRequestMsg,
    ResourceCallFunctionMsg,
    ResourceCallMethodMsg,
    ResourceCallSystemMethodMsg,
    ResourceAttrRequestMsg,
    ResourceHasAttrMsg,
    ResourceGetAttrMsg,
    ResourceSetAttrMsg,
    ResourceDelAttrMsg,
)

from .rpc_request_future import (
    FutureRequestMsg,
    CancelFutureMsg,
    CompleteFutureMsg,
)

from .rpc_event import (
    EventMsg,
    FutureEventMsg,
    FutureCanceledMsg,
    FutureCompletedMsg,
)

from .rpc_sideband import (
    SidebandMsg,
    ChannelMsg,
    RemotePrintMsg,
    FutureResultMsg,
    CHANNEL_CLOSED_MSG
)

################################################################################

from .rpc_result import (
    ResultMsg,
    ResultValueMsg,
    OkMsg,
    ValueMsg,
    JsonValueMsg,
    ResultErrorMsg,
    ErrorMsg,
    InternalErrorMsg,
    ResultUnavailableMsg,
    OK_MSG,
)

################################################################################

from .vars import VarsMsg, NO_VARS

################################################################################

from . import system as SYS

################################################################################

from .msg_aliases import (
    ArgsMsg,
    KwargsMsg,
    AnnotationsMsg,
    ResourcesRecordMsg,
    MethodsMsg,
    AttributesMsg,
    ClassesTupleMsg,
    OverloadsMsg,
)

################################################################################

from .__finalize import finalize_message_classes

################################################################################


__all__ = [
    'Msg',

    'ArgsMsg',
    'KwargsMsg',
    'AnnotationsMsg',
    'ResourcesRecordMsg',
    'MethodsMsg',
    'AttributesMsg',
    'ClassesTupleMsg',
    'OverloadsMsg',

    'EncoderP',
    'EncoderContextP',
    'DecoderP',
    'DecoderContextP',
    'CodecP',
    'NULL_CODEC',

    'BadMsg',
    'NO_MSG',
    'CORRUPT_MSG',
    'UNEXPECTED_MSG',
    'is_bad_msg',

    'TermMsg',
    'TermPassByValMsg',
    'TermPassByRefMsg',

    'AtomMsg',
    'NumberMsg',
    'SingletonMsg',
    'NoneMsg',
    'NotImplMsg',
    'EllipsisMsg',
    'ClosedMsg',
    'StrLikeMsg',
    'StrMsg',
    'BytesMsg',
    'ContainerMsg',
    'MappingMsg',
    'DictMsg',
    'SequenceMsg',
    'ListMsg',
    'TupleMsg',
    'SetMsg',
    'FrozenSetMsg',
    'SlotObjMsg',
    'ReduceObjMsg',
    'RegexPatternMsg',
    'RegexMatchMsg',
    'ClassUnionMsg',
    'EnumMemberMsg',
    'EnumKeyMsg',
    'EnumValMsg',
    'ExceptionMsg',
    'SyntaxMsg',
    'LambdaMsg',

    'ResourceMsg',
    'UserResourceMsg',
    'SystemResourceMsg',
    'RemoteResourceMsg',
    'LocalResourceMsg',
    'class_to_system_msg',
    'class_to_system_id',
    'pointer_to_system_msg',
    'pointer_to_system_id',

    'DefinitionMsg',
    'ResourceDataMsg',
    'ObjectDataMsg',
    'ClassDataMsg',
    'FunctionDataMsg',
    'CoroutineDataMsg',
    'ModuleDataMsg',
    'TypeDataMsg',
    'TypeAliasDataMsg',
    'TypeVarDataMsg',
    'GenericAliasDataMsg',
    'CallableTypeDataMsg',
    'TypeUnionDataMsg',
    'IteratorDataMsg',
    'EnumClassDataMsg',

    'RPCMsg',
    'FramedRequestMsg',
    'FramedResponseMsg',

    'RequestMsg',

    'ResourceRequestMsg',
    'ResourceCallRequestMsg',
    'ResourceNewMsg',
    'ResourceCallFunctionMsg',
    'ResourceCallMethodMsg',
    'ResourceCallSystemMethodMsg',
    'ResourceAttrRequestMsg',
    'ResourceHasAttrMsg',
    'ResourceGetAttrMsg',
    'ResourceSetAttrMsg',
    'ResourceDelAttrMsg',

    'FutureRequestMsg',
    'CancelFutureMsg',
    'CompleteFutureMsg',

    'EventMsg',
    'FutureEventMsg',
    'FutureCanceledMsg',
    'FutureCompletedMsg',

    'ReplRequestMsg',
    'ReplInitMsg',
    'ReplCallMethodMsg',
    'ReplRunCodeMsg',

    'SidebandMsg',
    'ChannelMsg',
    'RemotePrintMsg',
    'FutureResultMsg',
    'CHANNEL_CLOSED_MSG',

    'ResultMsg',
    'ResultValueMsg',
    'OkMsg',
    'ValueMsg',
    'JsonValueMsg',
    'ResultErrorMsg',
    'ErrorMsg',
    'InternalErrorMsg',
    'ResultUnavailableMsg',

    'NO_VARS',
    'OK_MSG',
    'SYS',

    'VarsMsg',

    'msgpack_to_rpc_msg',
    'msgpack_to_vars_msg',

    'rpc_msg_to_msgpack',
    'vars_msg_to_msgpack',

    'pprint_json',
    'pprint_msgpack',
    'fmt_json',
    'fmt_msgpack',

    'bytes_to_repl_init_data',

    'legacy',
]


################################################################################

finalize_message_classes()

################################################################################

msgpack_to_rpc_msg = RPCMsg.from_msgpack
msgpack_to_vars_msg = VarsMsg.from_msgpack

def rpc_msg_to_msgpack(msg: RPCMsg) -> bytes:
    assert isinstance(msg, RPCMsg)
    return msg.to_msgpack()

def vars_msg_to_msgpack(msg: VarsMsg) -> bytes:
    assert isinstance(msg, VarsMsg)
    return msg.to_msgpack()

################################################################################

def bytes_to_repl_init_data(globals_data: bytes, locals_data: bytes) -> tuple[ReplInitMsg,
tuple[DefinitionMsg, ...]]:

    globals_msg = VarsMsg.from_msgpack(globals_data)
    locals_msg = VarsMsg.from_msgpack(locals_data)

    init_msg = ReplInitMsg(globals_msg.vars, locals_msg.vars)
    def_msgs = globals_msg.defs + locals_msg.defs

    return init_msg, def_msgs



# fmt: off

import typing
from dataclasses import KW_ONLY
from enum import EnumType
from typing import _TypedDictMeta  # type: ignore
from typing import _ProtocolMeta

from ..msg.term_exception import EXC_NEW_ARGS
from .__ import *
from .base import *
from .handle import is_virtual_class
from .stub_methods import V_CLASS_GETITEM, V_OBJECT_METHODS

__all__ = [
    'ClassData',
]


################################################################################

class ClassData(ResourceData):
    __slots__ = (
        'name', 'qname', 'module', 'doc', 'cls', 'bases', 'methods', 'attrs',
        'sattrs', 'params', 'keys', 'annos',
    )

    FORBIDDEN_FORM = forbidden_class

    name:     str
    qname:    optstr
    module:   optstr
    doc:      optstr
    cls:      ClassT
    bases:    ClassesTupleT
    methods:  MethodsT
    attrs:    AttributesT
    sattrs:   strtup
    params:   Tup[TypeT]
    annos:    AnnotationsT
    keys:     strtup

    # attrs is for class-level attributes whose values ARE recreated locally
    # in the special case of cls = EnumMeta, these class-level attributes will include members
    # of the enumeration, and these are presented as serialized as `EnumMemberData` objects

    # `keys` is for any fields that getattr should work on but ARE NOT present in `methods` or `attrs`
    # these are represent attributes of the class object itself, in other words, if they have
    # corresponding annotations in `annos`, these annotations would be wrapped in ClassVar[...]

    # `sattrs` corresponds to the confusingly-named `__static_attributes__` field, which is what the
    # python compiler populates by finding all the `self.XXX` that occur in methods of a class
    # definition. we reproduce it here, but it is not semantically meaningful for virtualization.

    # implementation attached later
    @classmethod
    def describe_resource(cls, cls_obj: ClassT) -> 'ClassData': ...  # noqa: F841

    # implementation attached later
    def create_resource(self, handle: ResourceHandle) -> ClassT: ...


################################################################################

# TODO: is `__foo__` supposed to be considered private?
def is_private_field(ckey: str) -> bool:
    return ckey.startswith('_')

def describe_real_class(cls: type) -> ClassData:
    data = ClassData()

    assert isinstance(cls, type), f"{f_object_id(cls)} is not a class"

    bases, cdict, name, qname, module, doc = get_cls_attrs(cls)

    qname = qname if type(qname) is str else None
    doc = doc if type(doc) is str else None
    module = module if type(module) is str else None

    if not is_str(name):
        raise E.WarpEncodingError(f"class {cls!r} name {f_object_id(name)} is not a string")

    if not is_tup(bases):
        raise E.WarpEncodingError(f"class {cls!r} bases {f_object_id(bases)} is not a tuple")

    if is_forbidden(cls, module):
        raise E.WarpEncodingForbiddenError(f"<class {cls.__module__}.{cls.__qualname__}'>")

    data.name = name
    data.qname = qname
    data.module = module
    data.doc = doc

    mcls = type(cls)
    data.cls = mcls
    data.bases = bases

    no_priv_fields = flags.OMIT_PRIVATE_FIELDS
    no_priv_annos = flags.OMIT_PRIVATE_ANNOS
    allow_dunders = flags.ALLOW_KNOWN_DUNDER_METHODS

    data.methods, add_meth = mkdict()
    data.attrs, add_attr = mkdict()
    keys = []
    for ckey, cval in cdict.items():
        if ckey in CLS_IGNORE:
            continue
        if ckey.startswith('___'):
            continue
        if no_priv_fields and is_private_field(ckey) and ckey != '__init__':
            if not allow_cls_attr(ckey, allow_dunders):
                continue
        # TODO: we need to treat properties as first-class, this works for now because we have open keys but
        # stub generation is broken because it doesn't know about these properties on the classes.
        if is_property_t(cval):
            continue
        elif is_method_t(cval):
            add_meth(ckey, cval)
        elif ckey in CLASS_ATTR_WHITELIST:
            add_attr(ckey, cval)
        else:
            keys.append(ckey)

    data.keys = tuple(keys)
    data.params = getattr(cls, TPARAMS, ())

    sattrs, annos = multi_get_raw(cls, SATTRS, ANNOS)

    if not is_strtup(sattrs):
        sattrs = ()

    if not is_rec(annos):
        annos = {}

    if sattrs and no_priv_fields:
        sattrs = tuple(s for s in sattrs if not is_private_field(s))

    if annos and no_priv_annos:
        annos = {k: v for k, v in annos.items() if not is_private_field(k) or v is KW_ONLY}

    data.sattrs = sattrs
    data.annos = annos

    return data


################################################################################

def allow_cls_attr(ckey: str, allow_dunders: bool) -> bool:
    if not ckey.startswith('_'):
        return True
    elif ckey in CLASS_ATTR_WHITELIST:
        return True
    elif allow_dunders and ckey.endswith('__') and ckey in KNOWN_DUNDER_METHODS:
        return True
    else:
        return False

################################################################################

def create_virtual_class(data: ClassData, handle: ResourceHandle) -> ClassT:

    handle.kind = Kind.Class

    name = data.name
    qname = data.qname
    bases = data.bases

    cdict, add = mkdict()

    # REMOTE KEYS ==============================================================

    # here we decide what getattr keys should be allowed to result in a GetAttr request

    # the virtual class's keys = the union of what came over wire, and the keys of any virtual
    # parents
    keys = set(data.keys)

    any_virtual_bases = False
    for base in bases:
        if is_virtual_class(base):
            any_virtual_bases = True

            if base.__flags__ & HEAP_CLASS:
                keys |= cls_handle(base).keys

    if not any_virtual_bases:
        # if no bases at all, or none of the bases are virtual:
        # pick up the virtual versions of __str__, __hash__, etc; subclasses of this virtual
        # class will inherit these, or the true class may override them later in this
        # function, or in a subclass
        cdict.update(V_OBJECT_METHODS)

    # handle is the buried handle that tells us how to do virtual calls
    handle.keys = keys
    handle.open = flags.CLASS_OPEN_KEYS
    handle.name = name

    # OBJECT INSTANTIATION =====================================================

    # this `__new__` stub ensures that `v_cls(...)` will result in a `New` request, which will
    # construct an *ordinary* object remotely, whereupon it will be encoded to `ObjectDataMsg` when
    # the returned value is encoded. when this is decoded back here, it will trigger
    # `ObjectData.create_virtual_resource`, which will create an instance of the `v_cls` we are
    # defining here.

    # `__new__` is automatically class method, so when it is invoked by Python, the `cls`
    # argument isn't necessarily equal to `v_cls` (it could be a subclass of `v_cls`). but
    # because every such subclass must have been created by decoding, it will have its own
    # `__new__` that is more specialized. so we are guaranteed that `cls is v_cls`, though we
    # cannot assert this here because `v_cls` does not exist yet and so cannot be closed over.

    def __new__(cls, *args, **kwargs):
        if len(args) == 1 and type(args[0]) is ResourceHandle:
            base = object
            if issubclass(cls, Exception):
                base = Exception
            elif issubclass(cls, BaseException):
                base = BaseException
            v_obj = base.__new__(cls)
            obj_set(v_obj, VHDL, args[0])
            return v_obj
        v_obj = handle.hdlr(handle, ResourceNew(cls, args, kwargs))
        if issubclass(cls, (BaseException, Exception)):
            object.__setattr__(v_obj, EXC_NEW_ARGS, tuple(args))
        return v_obj

    # VIRTUAL METHODS ==========================================================

    # add methods
    for k, m in data.methods.items():
        add(k, m)

    add('__new__', __new__)

    # ATTRIBUTES ===============================================================

    # add (real) non-method attributes
    for k, a in data.attrs.items():
        add(k, a)

    # METADATA =================================================================

    # add metadata
    add(VHDL, handle)
    add(QUALNAME, qname)
    add(MODULE, data.module)
    add(DOC, data.doc)
    add(SATTRS, data.sattrs)
    add(ANNOS, data.annos)
    add(TPARAMS, data.params)

    # CLASS HIERARCHY ==========================================================

    bases = tuple(b for b in data.bases if b not in EXCLUDE_BASES)

    if data.params:
        # this allows (virtual) MyClass[int] to resolve to a typing._GenericAlias locally
        add('__class_getitem__', V_CLASS_GETITEM)

    mcls = data.cls

    # CLASS CREATION ===========================================================

    error: TypeError | None = None
    try:
        if mcls is _ProtocolMeta:
            cdict['_is_protocol'] = True
            cdict['_is_runtime_protocol'] = True

        if typing.Protocol in bases and any(not issubclass(b, typing.Protocol) for b in bases):
            bases = tuple(b for b in bases if b is not typing.Protocol)

        if mcls is EnumType:
            raise E.WarpDecodingError('enums not supported')
        elif mcls is _TypedDictMeta:
            bases = tuple(b for b in bases if b is not dict)
            # typeddict meta takes extra keyword argument `total`
            total = data.attrs.get('__total__', True)
            v_cls = mcls(name, bases, cdict, total=total)
        else:
            v_cls = mcls(name, bases, cdict)

    except Exception as exc:
        error = exc
        v_cls = type

    if error:
        f_error = D.fmt_exception(error)
        problem = f"Error creating virtual class {name!r} with bases {bases!r} and mcls {mcls!r}:\n"
        raise E.WarpDecodingError(problem + f_error)

    # CLASS ATTRIBUTES =========================================================

    # we set these AFTER class creation, so that _ProtocolMeta does not see these
    # attributes and getattr them, triggering RPC
    if flags.VIRTUAL_CLASS_ATTRIBUTES:
        # these will cause v_cls.foo to produce a ResourceGetAttr request against
        # v_cls. this is necessary because `v_cls.__getattr__` (which is really the `__getattr__` from
        # stub_methods.py) will only trigger for *instances* of v_cls.
        for key in keys:
            if key in cdict:
                # this shouldn't happen: keys is for attributes that aren't already
                # a method or an attribute
                continue
            v_attr = class_attribute(handle)
            cls_set(v_cls, key, v_attr)
            v_attr.__set_name__(v_cls, key)

    return v_cls


###############################################################################

class class_attribute:
    # ___vhdl___ is the handle for the class itself
    __slots__ = 'name', '__objclass__', VHDL

    def __init__(self, handle: ResourceHandle) -> None:
        self.___vhdl___ = handle
        self.__objclass__ = object
        self.name = '<unset>'

    def __set_name__(self, owner: type, name: str) -> None:
        self.__objclass__ = owner
        self.name = name

    def __get__(self, instance, owner: Any = None):
        handle = self.___vhdl___
        if instance is not None:
            # Instance access (e.g., v_obj.attr) - get from instance
            # This handles dataclass fields where the class has a default but
            # the instance has a different value
            from .handle import obj_handle
            inst_handle = obj_handle(instance)
            request = ResourceGetAttr(instance, self.name)
            return inst_handle.hdlr(inst_handle, request)
        else:
            # Class access (e.g., VirtualCls.attr) - get from class
            request = ResourceGetAttr(self.__objclass__, self.name)
            return handle.hdlr(handle, request)


    def __repr__(self):
        return f'<attribute {self.name!r}>'

    __str__ = __repr__


###############################################################################

HEAP_CLASS = 1 << 9

get_cls_attrs = O.attrgetter(BASES, DICT, NAME, QUALNAME, MODULE, DOC)

CLS_REQUIRED = (
    BASES,
    DICT,
    NAME,
    QUALNAME,
    MODULE,
    DOC,
)
CLS_OPTIONAL = SATTRS, ANNOS
CLS_INTERNAL = WEAKREF, FSTLINE, SLOTS
CLS_HOOKED = '__str__', '__repr__', '__hash__', '__getattr__', '__setattr__', '__delattr__'
CLS_IGNORE = CLS_REQUIRED + CLS_OPTIONAL + CLS_INTERNAL + CLS_HOOKED

EXCLUDE_BASES = [T.Generic]

################################################################################

# attach implementations to class
ClassData.describe_resource = staticmethod(describe_real_class)
ClassData.create_resource = create_virtual_class



# fmt: off

from typing import TYPE_CHECKING
from collections.abc import Callable, Coroutine


__all__ = [
    'SyncSendBytes',
    'SyncRecvBytes',
    'SyncRecvReady',
    'SyncWriteLog',
    'AsyncSendBytes',
    'AsyncRecvBytes',
    'AsyncSendRecvBytes',
    'AsyncSendMsg',
    'AsyncRecvMsg',
    'QUIT',
]


################################################################################

if TYPE_CHECKING:
    from ..msg.rpc import RPCMsg

################################################################################

type SyncSendBytes = Callable[[bytes], None]
type SyncRecvBytes = Callable[[], bytes]
type SyncRecvReady = Callable[[], bool]
type SyncWriteLog = Callable[[str], None]

type AsyncSendBytes = Callable[[bytes], Coroutine[None, None, None]]
type AsyncRecvBytes = Callable[[], Coroutine[None, None, bytes]]
type AsyncSendRecvBytes = Callable[[bytes], Coroutine[None, None, bytes]]

type AsyncSendMsg = Callable[['RPCMsg'], Coroutine[None, None, None]]
type AsyncRecvMsg = Callable[[], Coroutine[None, None, 'RPCMsg']]

QUIT: bytes = b'\0'



# fmt: off

from typing import Any, NamedTuple, TYPE_CHECKING

from .sentinels import Sentinel, is_sentinel, ERRORED, PENDING, CANCELED

__all__ = [
    'Result',
    'set_future_result',
    'get_future_result',
    'OK_RESULT',
    'PENDING_RESULT',
    'CANCELED_RESULT',
]

################################################################################

if TYPE_CHECKING:
    import asyncio

################################################################################

class Result(NamedTuple):
    done: bool
    value: Any
    error: BaseException | None

    @staticmethod
    def unavailable(sentinel: Sentinel | None = None) -> 'Result':
        if sentinel is PENDING:
            return PENDING_RESULT
        elif sentinel is ERRORED:
            return ERRORED_RESULT
        elif sentinel is CANCELED:
            return CANCELED_RESULT
        return Result(False, sentinel, None)

    @staticmethod
    def good(value: Any) -> 'Result':
        if value is None:
            return OK_RESULT
        return Result(True, value, None)

    @staticmethod
    def bad(exc: BaseException) -> 'Result':
        return Result(True, ERRORED, exc)

    @staticmethod
    def from_future(future: 'asyncio.Future') -> 'Result':
        return get_future_result(future)

    def into_future(self, future: 'asyncio.Future') -> None:
        set_future_result(future, self)

    @property
    def is_ok(self) -> bool:
        return self.done and self.error is None

    @property
    def is_err(self) -> bool:
        return self.done and self.error is not None

    @property
    def is_unavailable(self) -> bool:
        return not self.done

    @property
    def is_pending(self) -> bool:
        return not self.done and self.value is PENDING

    @property
    def is_canceled(self) -> bool:
        return not self.done and self.value is CANCELED

    @property
    def is_completed(self) -> bool:
        return self.done

    @property
    def sentinel(self) -> Sentinel | None:
        val = self.value
        return val if isinstance(val, Sentinel) else None

    def realize(self) -> Any:
        err = self.error
        val = self.value
        if not self.done:
            if is_sentinel(val):
                raise RuntimeError(f'result not available: {val}')
            raise RuntimeError('result not available')
        if err is not None:
            raise self.error
        return self.value

    def summary_str(self, width: int) -> str:
        from .fmt import f_object_id

        if self.is_err:
            f_error = f_object_id(self.error, width)
            return f'bad({f_error})'

        val = self.value

        if not self.done:
            return f'unavailable({val})'

        if type(val) is tuple:
            f_val = ', '.join(f_object_id(v, width//4) for v in val[:5])
            if len(val) > 5:
                f_val += ', ...'
            f_val = f'({f_val})'
        elif type(val) is str and len(val) <= width:
            f_val = repr(val)
        else:
            f_val = f_object_id(val, width)
        return f'good({f_val})'

    def __repr__(self) -> str:
        return self.summary_str(width=512)

    def __str__(self) -> str:
        return 'Result.' + self.summary_str(width=64)


################################################################################

def get_future_result(future: 'asyncio.Future') -> Result:
    if not future.done():
        return PENDING_RESULT
    elif future.cancelled():
        return CANCELED_RESULT
    elif (exception := future.exception()) is not None:
        return Result.bad(exception)
    else:
        return Result.good(future.result())

def set_future_result(future: 'asyncio.Future', result: Result) -> None:
    if result is PENDING_RESULT:
        return
    elif result is CANCELED_RESULT:
        future.cancel()
    elif result.is_err:
        future.set_exception(result.error)
    else:
        future.set_result(result.value)

###############################################################################

OK_RESULT       = Result(True, None, None)
PENDING_RESULT  = Result(False, PENDING, None)
CANCELED_RESULT = Result(False, CANCELED, None)
ERRORED_RESULT  = Result(False, ERRORED, None)

# NOTE: Result.unavailable(PENDING) will return the singleton PENDING_RESULT, etc.
# ERRORED_RESULT is for out-of-band errors, unlike Result.bad(exception) which is
# to represent a computation that raised an exception which we wish to represent.



import asyncio

from agentica_internal.repl.repl import *


def verify_sync_exec():
    repl = BaseRepl()
    ev = repl.run_code('''
    x = 1 + 1
    ''')
    assert not ev.error
    assert repl.vars.locals['x'] == 2


def verify_async_exec():
    repl = BaseRepl()
    loop = asyncio.new_event_loop()
    repl.set_loop(loop)
    ev = repl.run_code('''
    import asyncio
    await asyncio.sleep(0.1)
    ''')
    assert not ev.error
    assert 0.1 < ev.duration < 0.2


if __name__ == '__main__':
    verify_sync_exec()
    verify_async_exec()



################################################################################

from typing import overload, Any, Iterable, Literal
from collections.abc import Iterable

################################################################################

__classes__ = [
    '_SupportsCeil',
    '_SupportsFloor',
    '_SupportsProdWithNoDefaultGiven',
    '_SupportsTrunc',
]

__functions__ = [
    'acos',
    'acosh',
    'asin',
    'asinh',
    'atan',
    'atan2',
    'atanh',
    'ceil',
    'ceil',
    'comb',
    'copysign',
    'cos',
    'cosh',
    'degrees',
    'dist',
    'erf',
    'erfc',
    'exp',
    'expm1',
    'fabs',
    'factorial',
    'floor',
    'floor',
    'fmod',
    'frexp',
    'fsum',
    'gamma',
    'gcd',
    'hypot',
    'isclose',
    'isinf',
    'isfinite',
    'isnan',
    'isqrt',
    'lcm',
    'ldexp',
    'lgamma',
    'log',
    'log10',
    'log1p',
    'log2',
    'modf',
    'perm',
    'pow',
    'prod',
    'prod',
    'prod',
    'radians',
    'remainder',
    'sin',
    'sinh',
    'sqrt',
    'tan',
    'tanh',
    'trunc',
    'ulp',
]

__aliases__ = ['_SupportsFloatOrIndex', '_PositiveInteger', '_NegativeInteger']

__annos__ = []

__constants__ = []

__all__ = [
    'SupportsCeil',
    'SupportsFloor',
    'SupportsProdWithNoDefaultGiven',
    'SupportsTrunc',
    'acos',
    'acosh',
    'asin',
    'asinh',
    'atan',
    'atan2',
    'atanh',
    'ceil',
    'comb',
    'copysign',
    'cos',
    'cosh',
    'degrees',
    'dist',
    'erf',
    'erfc',
    'exp',
    'expm1',
    'fabs',
    'factorial',
    'floor',
    'fmod',
    'frexp',
    'fsum',
    'gamma',
    'gcd',
    'hypot',
    'isclose',
    'isinf',
    'isfinite',
    'isnan',
    'isqrt',
    'lcm',
    'ldexp',
    'lgamma',
    'log',
    'log10',
    'log1p',
    'log2',
    'modf',
    'perm',
    'pow',
    'prod',
    'radians',
    'remainder',
    'sin',
    'sinh',
    'sqrt',
    'tan',
    'tanh',
    'trunc',
    'ulp',
]

################################################################################


class _SupportsCeil:
    def __ceil__(self) -> Any: ...


################################################################################


class _SupportsFloor:
    def __floor__(self) -> Any: ...


################################################################################


class _SupportsProdWithNoDefaultGiven:
    pass


################################################################################


class _SupportsTrunc:
    def __trunc__(self) -> Any: ...


################################################################################


def acos(x: float | int, /) -> float: ...


def acosh(x: float | int, /) -> float: ...


def asin(x: float | int, /) -> float: ...


def asinh(x: float | int, /) -> float: ...


def atan(x: float | int, /) -> float: ...


def atan2(y: float | int, x: float | int, /) -> float: ...


def atanh(x: float | int, /) -> float: ...


@overload
# pyrefly: ignore[invalid-overload]
def ceil(x: _SupportsCeil, /) -> Any: ...


@overload
# pyrefly: ignore[invalid-overload]
def ceil(x: float | int, /) -> int: ...


def comb(n: int, k: int, /) -> int: ...


def copysign(x: float | int, y: float | int, /) -> float: ...


def cos(x: float | int, /) -> float: ...


def cosh(x: float | int, /) -> float: ...


def degrees(x: float | int, /) -> float: ...


def dist(p: Iterable[float | int], q: Iterable[float | int], /) -> float: ...


def erf(x: float | int, /) -> float: ...


def erfc(x: float | int, /) -> float: ...


def exp(x: float | int, /) -> float: ...


def expm1(x: float | int, /) -> float: ...


def fabs(x: float | int, /) -> float: ...


def factorial(x: int, /) -> int: ...


@overload
# pyrefly: ignore[invalid-overload]
def floor(x: _SupportsFloor, /) -> Any: ...


@overload
# pyrefly: ignore[invalid-overload]
def floor(x: float | int, /) -> int: ...


def fmod(x: float | int, y: float | int, /) -> float: ...


def frexp(x: float | int, /) -> tuple: ...


def fsum(seq: Iterable[float | int], /) -> float: ...


def gamma(x: float | int, /) -> float: ...


def gcd(*integers: int) -> int: ...


def hypot(*coordinates: float | int) -> float: ...


def isclose(
    a: float | int, b: float | int, *, rel_tol: float | int = 1e-09, abs_tol: float | int = 0.0
) -> bool: ...


def isinf(x: float | int, /) -> bool: ...


def isfinite(x: float | int, /) -> bool: ...


def isnan(x: float | int, /) -> bool: ...


def isqrt(n: int, /) -> int: ...


def lcm(*integers: int) -> int: ...


def ldexp(x: float | int, i: int, /) -> float: ...


def lgamma(x: float | int, /) -> float: ...


def log(x: float | int, base: float | int = ...) -> float: ...


def log10(x: float | int, /) -> float: ...


def log1p(x: float | int, /) -> float: ...


def log2(x: float | int, /) -> float: ...


def modf(x: float | int, /) -> tuple: ...


def perm(n: int, k: int | None = None, /) -> int: ...


def pow(x: float | int, y: float | int, /) -> float: ...


@overload
# pyrefly: ignore[invalid-overload]
def prod(iterable: Iterable[bool | int], /, *, start: int = 1) -> int: ...


@overload
# pyrefly: ignore[invalid-overload]
def prod(iterable: Iterable, /) -> Any | Literal[1]: ...


@overload
# pyrefly: ignore[invalid-overload]
def prod(iterable: Iterable, /, *, start: Any) -> Any: ...


def radians(x: float | int, /) -> float: ...


def remainder(x: float | int, y: float | int, /) -> float: ...


def sin(x: float | int, /) -> float: ...


def sinh(x: float | int, /) -> float: ...


def sqrt(x: float | int, /) -> float: ...


def tan(x: float | int, /) -> float: ...


def tanh(x: float | int, /) -> float: ...


def trunc(x: _SupportsTrunc, /) -> Any: ...


def ulp(x: float | int, /) -> float: ...



from agentica_internal.core.sentinels import ARG_DEFAULT
from agentica_internal.warpc.flags import with_flags
from agentica_internal.warpc.worlds.debug_world import DebugWorld

my_object = object()


def my_func(a: int = 0, b: str = 'hello', c: object = my_object):
    return a, b, c


async def verify_no_func_defaults():
    with with_flags(VIRTUAL_FUNCTION_DEFAULTS=None):
        pair = DebugWorld.connected_pair()
        async with pair:
            A, B = pair.pipes

            my_func_b = await B(my_func)
            assert my_func_b.__defaults__ == (ARG_DEFAULT, ARG_DEFAULT, ARG_DEFAULT)

            result_b = my_func_b()
            result_a = await A(result_b)
            assert result_a == (0, 'hello', my_object)


async def verify_atom_func_defaults():
    with with_flags(VIRTUAL_FUNCTION_DEFAULTS='atoms'):
        pair = DebugWorld.connected_pair()
        async with pair:
            A, B = pair.pipes

            my_func_b = await B(my_func)
            assert my_func_b.__defaults__ == (0, 'hello', ARG_DEFAULT)

            result_b = my_func_b()
            result_a = await A(result_b)
            assert result_a == (0, 'hello', my_object)


async def verify_all_func_defaults():
    with with_flags(VIRTUAL_FUNCTION_DEFAULTS='atoms'):
        pair = DebugWorld.connected_pair()
        async with pair:
            A, B = pair.pipes

            my_func_b = await B(my_func)
            assert len(my_func_b.__defaults__) == 3

            result_b = my_func_b()
            result_a = await A(result_b)
            assert result_a == (0, 'hello', my_object)


async def main():
    await verify_no_func_defaults()
    await verify_atom_func_defaults()
    await verify_all_func_defaults()


if __name__ == '__main__':
    import asyncio

    asyncio.run(main())



from agentica_internal.repl.repl import *


def verify_stdout_capture():
    repl = BaseRepl()
    data = repl.run_code('''
    import sys
    import io
    sys.stdout = buffer = io.StringIO()
    print('hello')
    captured = buffer.getvalue()
    ''')
    assert not data.has_error
    captured = repl.vars.locals['captured']
    assert isinstance(captured, str)
    assert captured == 'hello\n'


if __name__ == '__main__':
    verify_stdout_capture()



# fmt: off

from typing import Any
from builtins import type as cls_t, object as obj_t
from types import ModuleType as mod_t

from .handle import ResourceHandle

__all__ = [
    'cls_get',  'obj_get',  'mod_get',
    'cls_set',  'obj_set',  'mod_set',
    'cls_handle', 'mod_handle', 'obj_handle',
    'cls_dict', 'mod_dict', 'obj_dict',
]


################################################################################

def cls_handle(cls: cls_t) -> ResourceHandle: ...
def mod_handle(mod: mod_t) -> ResourceHandle: ...
def obj_handle(cls: obj_t) -> ResourceHandle: ...

def cls_dict(cls: cls_t) -> dict[str, Any]: ...
def mod_dict(mod: mod_t) -> dict[str, Any]: ...
def obj_dict(cls: obj_t) -> dict[str, Any]: ...

def cls_get(cls: cls_t, key: str) -> Any: ...
def mod_get(mod: mod_t, key: str) -> Any: ...
def obj_get(cls: obj_t, key: str) -> Any: ...

def cls_set(cls: cls_t, key: str, val: Any) -> None: ...
def mod_set(mod: mod_t, key: str, val: Any) -> None: ...
def obj_set(cls: obj_t, key: str, val: Any) -> None: ...



from agentica_internal.cpython.shed.load import get_shed_function, get_shed_method, get_shed_module


def verify_builtins_shed():
    module = get_shed_module('builtins')
    assert 'str' in getattr(module, '__all__', ())
    method = get_shed_method(str, 'capitalize')
    assert method.__qualname__ == 'str.capitalize'
    assert method.__module__ == 'builtins'
    assert method.__annotations__.get('return') == str


def verify_math_shed():
    module = get_shed_module('math')
    assert 'sqrt' in getattr(module, '__all__', ())
    fn = get_shed_function('math', 'sqrt')
    assert fn.__module__ == 'math'
    assert fn.__name__ == 'sqrt'
    assert fn.__annotations__.get('return') == float


def verify_not_in_shed():
    class MyClass: ...

    MyClass.__module__ = 'my_module'
    method = get_shed_method(MyClass, 'foo')
    assert method.__module__ == 'my_module'
    assert method.__name__ == 'foo'
    assert method.__qualname__ == 'MyClass.foo'
    assert not method.__annotations__

    method = get_shed_function('my_module', 'foo')
    assert method.__module__ == 'my_module'
    assert method.__name__ == 'foo'
    assert method.__qualname__ == 'foo'
    assert not method.__annotations__


if __name__ == '__main__':
    verify_builtins_shed()
    verify_math_shed()
    verify_not_in_shed()



from collections.abc import Iterable, Iterator
from typing import Self, cast


class chaindict[K, V](dict[K, V]):
    chain: list[dict[K, V]]
    _key_cache: list[K] | None

    __sentinel = object()

    def __init__(self, *chain: dict[K, V]):
        # only pretends to be a `dict`, but super().items() should be empty.
        super().__init__()
        self.chain = []
        self._key_cache = None

        for d in chain:
            if isinstance(d, chaindict):
                self.chain.extend(d.chain)
            else:
                self.chain.append(d)

    def __getitem__(self, key: K) -> V:
        for d in self.chain:
            if key in d:
                return d[key]
        raise KeyError(key)

    def __setitem__(self, key: K, value: V) -> None:
        if not self.chain:
            self.chain.append(dict())
        self.chain[0][key] = value
        self._key_cache = None

    def __delitem__(self, key: K) -> None:
        for d in self.chain:
            if key in d:
                del d[key]
                self._key_cache = None
                return
        raise KeyError(key)

    def __contains__(self, key: K) -> bool:
        return any(key in d for d in self.chain)

    def pop(self, key: K, default: V | object = __sentinel) -> V:
        for d in self.chain:
            if key in d:
                self._key_cache = None
                return d.pop(key)

        if default is self.__sentinel:
            raise KeyError(key)
        else:
            return cast(V, default)

    def copy(self) -> Self:
        """copy all chains, all referenced dictionaries are copied"""
        return type(self)(*(link.copy() for link in self.chain))

    def get[T](self, key: K, default: T = None) -> V | T:
        try:
            return self[key]
        except KeyError:
            return default

    def append_chain(self, link: dict[K, V]) -> None:
        if isinstance(link, chaindict):
            self.chain.extend(link.chain)
        else:
            self.chain.append(link)
        self._key_cache = None

    def prepend_chain(self, link: dict[K, V]):
        if isinstance(link, chaindict):
            self.chain[0:0] = link.chain
        else:
            self.chain.insert(0, link)
        self._key_cache = None

    def extend_chain(self, chain: Iterable[dict[K, V]]) -> None:
        for link in chain:
            self.append_chain(link)

    def pop_chain(self, index: int = -1) -> dict[K, V]:
        return self.chain.pop(index)

    def _compute_keys(self) -> list[K]:
        seen: set[K] = set()
        keys: list[K] = []

        for d in self.chain:
            for k in d:
                if k not in seen:
                    seen.add(k)
                    keys.append(k)

        return keys

    def keys(self) -> list[K]:
        if self._key_cache is None:
            self._key_cache = self._compute_keys()
        return list(self._key_cache)

    def values(self) -> list[V]:
        return [self[k] for k in self.keys()]

    def items(self) -> list[tuple[K, V]]:
        return [(k, self[k]) for k in self.keys()]

    def __len__(self) -> int:
        return len(self.keys())

    def flatten(self) -> dict[K, V]:
        return dict(self.items())

    def __iter__(self) -> Iterator[K]:
        return iter(self.keys())

    def __repr__(self):
        return f"{type(self).__name__}{tuple(self.chain)!r}"

    def __eq__(self, other: object) -> bool:
        if isinstance(other, chaindict):
            return self.flatten() == other.flatten()
        if isinstance(other, dict):
            return self.flatten() == other
        return NotImplemented

    @staticmethod
    def top(dct: dict[K, V]) -> dict[K, V] | None:
        """get the top-most dictionary in the chain if dct is a chaindict"""
        if isinstance(dct, chaindict):
            if dct.chain:
                return dct.chain[0]
        return None



# fmt: off

import asyncio
from typing import Any
from inspect import iscoroutinefunction

from agentica_internal.warpc.flags import with_flags
from agentica_internal.warpc.worlds.debug_world import DebugWorld, is_virtual


async def verify_coro_function():

    async def afunc(i: int) -> Any:
        return i * 2

    assert iscoroutinefunction(afunc)

    with with_flags(DEFAULT_ASYNC_MODE='coro'):

        pair = DebugWorld.connected_pair()
        async with pair:

            A, B = pair.pipes

            afunc_b = await B(afunc)
            assert is_virtual(afunc_b)
            assert iscoroutinefunction(afunc_b)

            afunc_a = await A(afunc_b)
            assert afunc_a is afunc

            coro_a = afunc_a(5)
            coro_b = afunc_b(5)

            assert coro_a.cr_code.co_filename == __file__
            assert coro_b.cr_code.co_filename == 'virtual'

            result_a = await coro_a
            result_b = await coro_b
            assert result_a == result_b, f"{result_a=!r} != {result_b=!r}"

        return pair


async def verify_coro_functions_run_concurrently():
    from time import time_ns

    async def sleepy(i: int) -> Any:
        await asyncio.sleep(1)
        return i * 2

    assert iscoroutinefunction(sleepy)

    with with_flags(DEFAULT_ASYNC_MODE='coro'):

        pair = DebugWorld.connected_pair()
        async with pair:

            A, B = pair.pipes

            sleepy_b = await B(sleepy)
            assert is_virtual(sleepy_b)
            assert iscoroutinefunction(sleepy_b)

            time_0 = time_ns()
            sleepies = tuple(sleepy_b(i) for i in range(10))
            slumber_party = asyncio.gather(*sleepies)
            results = await slumber_party
            time_1 = time_ns()

            assert results == [i*2 for i in range(10)]
            seconds = (time_1 - time_0) // 1e9
            assert 0.5 <= seconds <= 1.5

        return pair


def collect_async_events(events: list):
    asyncio.run(verify_coro_function()).collect_events(events)
    asyncio.run(verify_coro_functions_run_concurrently()).collect_events(events)


if __name__ == '__main__':
    asyncio.run(verify_coro_function())
    asyncio.run(verify_coro_functions_run_concurrently())



from enum import StrEnum
from typing import Literal

from msgspec import Struct, field, json, msgpack

from ..core.datetime import time_now_utc
from ..core.utils import subclasses_dict
from ..internal_errors import GenerationError, InvocationError

####################################################################################################

type AgenticaMsg = (
    InvocationWarpMsg
    | InvocationErrorMsg
    | InvocationEventMsg
    | InvocationCancelMsg
    | RequestInvocationMsg
    | InvocationCreatedMsg
)

type ClientMsg = RequestInvocationMsg | InvocationWarpMsg | InvocationCancelMsg

type ClientInvocationMsg = InvocationWarpMsg | InvocationCancelMsg

type ServerMsg = InvocationCreatedMsg | InvocationErrorMsg | InvocationWarpMsg | InvocationEventMsg

type ServerInvocationMsg = InvocationErrorMsg | InvocationWarpMsg | InvocationEventMsg

####################################################################################################

type AgenticaInvocationEventType = Literal['ENTER', 'EXIT', 'ERROR']

####################################################################################################

type InternalError = InvocationError | GenerationError

INTERNAL_ERROR_ClASSES: dict[str, type[Exception]]
INTERNAL_ERROR_ClASSES = subclasses_dict(GenerationError) | subclasses_dict(GenerationError)
INTERNAL_ERROR_ClASSES['RuntimeError'] = RuntimeError  # for unknown errors
INTERNAL_ERROR_NAMES = list(INTERNAL_ERROR_ClASSES.keys())

InternalErrorName = StrEnum("AgenticaErrorType", INTERNAL_ERROR_NAMES)

####################################################################################################


def make_timestamp() -> str:
    return time_now_utc()


timestamp_field = field(default_factory=make_timestamp)


####################################################################################################

# Triggering events from the client


class RequestInvocationMsg(Struct, tag="invoke"):
    match_id: str
    warp_locals_payload: bytes
    has_return_type: bool
    streaming: bool
    prompt: str | None = None
    timestamp: str = timestamp_field


####################################################################################################


class InvocationCancelMsg(Struct, tag="cancel"):
    iid: str
    timestamp: str = timestamp_field


####################################################################################################


# Generic carrier
class InvocationWarpMsg(Struct, tag="data"):
    iid: str
    data: bytes
    timestamp: str = timestamp_field


####################################################################################################


# Response from the server
class InvocationCreatedMsg(Struct, tag="new_iid"):
    iid: str
    match_id: str
    timestamp: str = timestamp_field


####################################################################################################


class InvocationErrorMsg(Struct, tag="error"):
    iid: str
    error_name: str  # this could be InternalErrorName, but this lets us send other exceptions
    error_message: str | None = None
    timestamp: str = timestamp_field

    @classmethod
    def from_error(cls, iid: str, err: GenerationError) -> "InvocationErrorMsg":
        err.__module__ = ""
        return cls(
            iid=iid,
            error_name=err.__class__.__name__,
            error_message=str(err),
        )

    def to_exception(self) -> Exception:
        error_name = self.error_name
        error_msg = self.error_message
        error_cls = INTERNAL_ERROR_ClASSES.get(error_name)
        if error_cls is None:
            error_cls = RuntimeError
            error_msg = f'{error_name}: {error_msg}'
        if error_msg:
            return error_cls(error_msg)
        return error_cls()


####################################################################################################


class InvocationEventMsg(Struct, tag="invocation"):
    iid: str
    event: AgenticaInvocationEventType
    timestamp: str = timestamp_field


####################################################################################################

_enc_json = json.Encoder()
_dec_json = json.Decoder(AgenticaMsg)

_enc_msgpack = msgpack.Encoder()
_dec_msgpack = msgpack.Decoder(AgenticaMsg)


def multiplex_from_json(json_bytes: bytes) -> AgenticaMsg:
    return _dec_json.decode(json_bytes)


def multiplex_to_json(message: AgenticaMsg) -> bytes:
    return _enc_json.encode(message)


def multiplex_to_msgpack(message: AgenticaMsg) -> bytes:
    return _enc_msgpack.encode(message)


def multiplex_from_msgpack(json_bytes: bytes) -> AgenticaMsg:
    return _dec_msgpack.decode(json_bytes)



# fmt: off

import typing
from collections.abc import Callable, Iterable
from re import compile as re_compile

__all__ = [
    'maptuple',
    'maplist',
    'mapdict',
    'identity',
    'constant_',
    'then_',
    'is_',
    'eq_',
    'neq_',
    'and_',
    'or_',
    'not_',
    'any_',
    'all_',
    'none_',
    'type_',
    'instance_',
    'in_',
    'prefixed_',
    'suffixed_',
    're_match_',
    'glob_match_',
    'true_fn',
    'false_fn',
    'none_fn',
]


################################################################################

AS = typing.ParamSpec('AS')
A = typing.TypeVar('A')
B = typing.TypeVar('B')
C = typing.TypeVar('C')
K = typing.TypeVar('K')


def maptuple(fn: Callable[[A], B], seq: Iterable[A]) -> tuple[B, ...]:
    return tuple(map(fn, seq))


def maplist(fn: Callable[[A], B], seq: Iterable[A]) -> list[B]:
    return list(map(fn, seq))


def mapdict(fn: Callable[[A], B], dct: dict[K, A]) -> dict[K, B]:
    return {k: fn(v) for k, v in dct.items()}


################################################################################


def identity(value: A) -> A:
    """`identity(val)` returns `val`"""
    return value


identity = typing._idfunc  # type: ignore


def constant_(value: A) -> Callable[..., A]:
    """`constant_(val)` returns `... => val`"""

    def fn(*_, **__):
        return value

    return fn


def then_(f: Callable[AS, B], g: Callable[[B], C]) -> Callable[AS, C]:
    """`then_(f, g)` returns `args.. => g(f(args..))`"""
    if f is identity:
        return g
    if g is identity:
        return f

    def fn(*_, **__):
        return g(f(*_, **__))

    return fn


###############################################################################

# predicates


def is_(a: A) -> A:
    """`is_(a)` returns `b => a is b`"""

    def fn(b):
        return a is b

    return fn


def eq_(a: A) -> A:
    """`eq_(a)` returns `arg => arg == a`"""

    def fn(b):
        return a == b

    return fn


def neq_(a: A) -> A:
    """`neq_(a)` returns `arg => arg != a`"""

    def fn(b):
        return a != b

    return fn


def and_(p: Callable[AS, bool], q: Callable[AS, bool]) -> Callable[AS, bool]:
    """`and_(p, q)` returns `args.. => p(args..) and q(args..)`"""

    def fn(*_, **__):
        return p(*_, **__) and q(*_, **__)

    return fn


def or_(p: Callable[AS, bool], q: Callable[AS, bool]) -> Callable[AS, bool]:
    """`or_(p, q)` returns `args.. => p(args..) or q(args..)`"""

    def fn(*_, **__):
        return p(*_, **__) or q(*_, **__)

    return fn


def not_(p: Callable[AS, bool]) -> Callable[AS, bool]:
    """`not_(p)` returns `args.. => not_fn(p(args..))`"""

    def fn(*_, **__):
        return not p(*_, **__)

    return fn


###############################################################################


def any_(p: Callable[[A], bool]) -> Callable[[Iterable[A]], bool]:
    """`any_(p)` returns `seq => any(p(a) for a in seq)`"""

    def fn(seq):
        return any(map(p, seq))

    return fn


def all_(p: Callable[[A], bool]) -> Callable[[Iterable[A]], bool]:
    """`all_(p)` returns `seq => all(p(a) for a in seq)`"""

    def fn(seq):
        return all(map(p, seq))

    return fn


def none_(p: Callable[[A], bool]) -> Callable[[Iterable[A]], bool]:
    """`none_(p)` returns `seq => not any(p(a) for a in seq)`"""

    def fn(seq):
        return not any(map(p, seq))

    return fn


###############################################################################


def type_(t: type | tuple[type, ...]) -> Callable[[object], bool]:
    """`type_(t)` returns `obj => type(obj) is t`"""
    if type(t) is tuple:

        def fn(obj):
            return type(obj) in t

        return fn

    def fn(obj):
        return type(obj) is t

    return fn


def instance_(t: type | tuple[type, ...]) -> Callable[[object], bool]:
    """`instance_(typ)` returns `obj => instanceof(obj, ty)`"""

    def fn(obj):
        return isinstance(obj, t)

    return fn


def in_(it: Iterable[A]) -> Callable[[A], bool]:
    """`in_(it)` returns `arg => arg in it`"""
    if not isinstance(it, (tuple, set, list, dict, frozenset)):
        it = tuple(it)
    return it.__contains__


###############################################################################


def prefixed_(prefix: str) -> Callable[[str], bool]:
    def fn(s: str) -> bool:
        return s.startswith(prefix)

    return fn


def suffixed_(suffix: str) -> Callable[[str], bool]:
    def fn(s: str) -> bool:
        return s.endswith(suffix)

    return fn


def re_match_(patt: str) -> Callable[[str], bool]:
    regex = re_compile(patt)

    def fn(s: str) -> bool:
        return bool(regex.fullmatch(s))

    return fn


def glob_match_(patt: str | None, default: bool = True) -> Callable[[str], bool]:
    if patt is None:
        return true_fn if default else false_fn

    assert type(patt) is str

    if '*' not in patt:
        return patt.__eq__

    if patt.count('*') == 1:
        if patt.startswith('*'):
            return suffixed_(patt[1:])
        if patt.endswith('*'):
            return prefixed_(patt[:-1])

    import fnmatch

    return re_match_(fnmatch.translate(patt))


###############################################################################

true_fn = constant_(True)
false_fn = constant_(False)
none_fn = constant_(None)



# fmt: off

from types import ModuleType, NoneType
from collections.abc import Callable
from typing import Any, TypeAliasType, ForwardRef, Union

import msgspec
from .resource_def import DefinitionMsg

from ...core.debug import avoid_file
from ...core.anno import union_iter
from ...core.print import eprint
from ...core.log import LogFlag

from ..alias import Tup, Rec, MethodKind
from .. import flags

from . import (
    base, bad,
    resource_data, resource_def,
    rpc, rpc_framed,
    rpc_request, rpc_request_repl, rpc_request_resource,
    rpc_result, rpc_sideband,
    term, term_atom, term_container, term_exception,
    term_resource, term_special
)

from .__json import fmt_json
from .__msgpack import pprint_msgpack

from .bad import NO_MSG, Msg
from .term import TermMsg, SIMPLE_ENCODERS, COMPLEX_ENCODERS, CONSTANT_ENCODERS
from .term_resource import ResourceMsg
from .resource_data import FunctionDataMsg
from .term_exception import ExceptionMsg
from .term_lambda import SyntaxMsg

__all__ = [
    'finalize_message_classes',
]


################################################################################

LOG_FINALIZE = False

avoid_file(__file__)

################################################################################

MSG_CLS_MODULES: list[ModuleType] = [
    base, bad,
    resource_data, resource_def,
    rpc, rpc_framed,
    rpc_request, rpc_request_repl, rpc_request_resource,
    rpc_result, rpc_sideband,
    term, term_atom, term_container, term_exception,
    term_resource, term_special
]

def _import_msg_classes():

    for mod in MSG_CLS_MODULES:
        imports = NAME_TO_MSG_CLS.copy()
        mod.__dict__.update(NAME_TO_MSG_CLS)


MSG_CLS_LIST:    list[type[Msg]] = []
NAME_TO_MSG_CLS: dict[str, type[Msg]] = {}

################################################################################

def _scan_msg_classes(fn: Callable[[type[Msg]], None]):

    for msg_cls in MSG_CLS_LIST:
        try:
            fn(msg_cls)
        except Exception as exc:
            exc.add_note(f'while applying {fn.__name__} to {msg_cls.__name__}')
            raise

################################################################################

upd_cnst = CONSTANT_ENCODERS.update
upd_atom = SIMPLE_ENCODERS.update
upd_comp = COMPLEX_ENCODERS.update

def _visit_cls(cls: type[Msg]):
    MSG_CLS_LIST.append(cls)
    NAME_TO_MSG_CLS[cls.__name__] = cls

    cls.LOG = LogFlag(cls)  # type: ignore
    subs = cls.__subclasses__()
    cls.TAG = cls.__struct_config__.tag

    if not subs:
        if hasattr(cls, 'V') and not issubclass(cls, ResourceMsg):
            types = tuple(union_iter(cls.V.__value__, type))
            if encode_fn := getattr(cls, 'encode_constant', None):
                msg = encode_fn()
                upd_cnst(dict.fromkeys(types, msg))
            if encode_fn := getattr(cls, 'encode_atom', None):
                upd_atom(dict.fromkeys(types, encode_fn))
            elif encode_fn := getattr(cls, 'encode_compound', None):
                upd_comp(dict.fromkeys(types, encode_fn))
        cls.LEAF_CLASSES = (cls,)
        cls.UNION = cls  # type: ignore
        return

    leafs = []
    extend = leafs.extend
    for sub in subs:
        _visit_cls(sub)
        extend(sub.LEAF_CLASSES)

    if cls is TermMsg and flags.INLINE_ATOMS:
        leafs.extend(flags.INLINE_ATOMS)

    if cls is ResourceMsg and flags.INLINE_DEFINITIONS:
        leafs.append(DefinitionMsg)

    cls.LEAF_CLASSES = tuple(leafs)

    union: Any = Union[*leafs]
    cls.UNION = union  # type: ignore


################################################################################

def _resolve_forward_refs(msg_cls: type[Msg]) -> None:

    annos = getattr(msg_cls, '__annotations__', None)
    if annos is None:
        return

    for key, anno in annos.items():
        if isinstance(anno, ForwardRef):
            anno = anno.__forward_arg__
        if isinstance(anno, TypeAliasType):
            anno = anno.__name__
        if isinstance(anno, str):
            try:
                parsed = parse(anno)
                annos[key] = parsed
            except:
                pass

def parse(anno: str):
    if equiv := get_equiv(anno):
        return equiv
    if anno.startswith('Tup['):
        parsed = parse(anno[4:-1])
        return Tup[parsed]
    elif anno.startswith('Rec['):
        parsed = parse(anno[4:-1])
        return Rec[parsed]
    elif ' | ' in anno:
        parsed = list(map(parse, anno.split(' | ')))
        return Union[*parsed]
    elif anno == 'None':
        return None
    cls = NAME_TO_MSG_CLS[anno]
    return cls.UNION


################################################################################

def _attach_decoders(cls: type[Msg]):
    cls_name = cls.__name__

    union_t = cls.UNION
    if LOG_FINALIZE:
        from ...core.type import anno_str
        if type(union_t) is not type(Msg):
            eprint(f'{cls_name:<20}{anno_str(union_t)}')
    try:
        json_decoder = msgspec.json.Decoder(union_t).decode
        msgpack_decoder = msgspec.msgpack.Decoder(union_t).decode
    except Exception as e:
        e.add_note(f'error while attaching decoders to {cls_name}:')
        f_annos = '\n'.join(f"\t{k:<20}: {v} {type(v)}" for k, v in cls.__annotations__.items())
        e.add_note(f_annos)
        raise

    def from_json(msg_cls, json_data: bytes):
        try:
            assert msg_cls is cls
            assert isinstance(json_data, bytes)
            json_data = bytes(json_data)
            return json_decoder(json_data)
        except Exception as exc:
            f_json = fmt_json(json_data)
            f_cls = msg_cls.__module__ + '.' + msg_cls.__name__
            exc.add_note(f'error while decoding {f_cls!r}:\n{f_json}')
            raise

    def from_msgpack(msg_cls, msgpack_data: bytes):
        if not isinstance(msgpack_data, bytes):
            return NO_MSG
        assert msg_cls is cls
        try:
            if not msgpack_data:
                return cls()
            return msgpack_decoder(msgpack_data)
        except Exception as exc:
            print_msgspec_error(cls_name, msgpack_data, exc)
            raise

    from_msgpack.__name__ = from_msgpack.__qualname__ = f'{cls_name}.from_msgpack'
    from_json.__name__ = from_json.__qualname__ = f'{cls_name}.from_json'

    cls.from_json = classmethod(from_json)
    cls.from_msgpack = classmethod(from_msgpack)


def print_msgspec_error(cls_name: str, msgpack_data: bytes, exc: Exception) -> None:
    eprint(
        f'\n\nERROR DECODING {cls_name}',
        'EXCEPTION:',
        exc,
        'MESSAGE FOLLOWS:',
        sep='\n',
    )
    pprint_msgpack(msgpack_data, err=True)


################################################################################

def _dispatch_on_sys_exc_types():
    from ..system import SYS_EXCEPTIONS
    upd_comp(dict.fromkeys(SYS_EXCEPTIONS, ExceptionMsg.encode_compound))


################################################################################

FINALIZED = False

EQUIVS: dict[str, Any] = {
    'str': str,
    'int': int,
    'float': float,
    'bool': bool,
    'None': NoneType
}
get_equiv = EQUIVS.get

def finalize_message_classes():
    eprint('\nfinalize_message_classes\n') if LOG_FINALIZE else None
    global FINALIZED
    if FINALIZED:
        eprint("warning: finalize_message_classes called more than once")
        return

    _visit_cls(Msg)
    MSG_CLS_LIST.reverse()
    _import_msg_classes()

    synt_msg_union = SyntaxMsg.UNION                           # type: ignore
    term_msg_union = TermMsg.UNION                             # type: ignore
    rsrc_msg_union = ResourceMsg.UNION                         # type: ignore
    term_tup_t = tuple[term_msg_union, ...]                    # type: ignore
    term_rec_t = dict[str, term_msg_union]                     # type: ignore
    rsrc_tup_t = tuple[rsrc_msg_union, ...]                    # type: ignore
    rsrc_rec_t = dict[str, rsrc_msg_union]                     # type: ignore
    synt_lst_t = list[synt_msg_union]                          # type: ignore
    method_pair_t = tuple[MethodKind, rsrc_msg_union]          # type: ignore
    methods_msg_t = dict[str, method_pair_t | rsrc_msg_union]  # type: ignore
    fn_data_tup_t: Any = tuple[FunctionDataMsg, ...]           # type: ignore

    EQUIVS['TermMsg']            = term_msg_union
    EQUIVS['SyntaxMsg']          = synt_msg_union
    EQUIVS['ResourceMsg']        = rsrc_msg_union
    EQUIVS['ArgsMsg']            = term_tup_t
    EQUIVS['KwargsMsg']          = term_rec_t
    EQUIVS['AnnotationsMsg']     = term_rec_t
    EQUIVS['AttributesMsg']      = term_rec_t
    EQUIVS['ClassesTupleMsg']    = rsrc_tup_t
    EQUIVS['OverloadsMsg']       = fn_data_tup_t
    EQUIVS['ResourcesRecordMsg'] = rsrc_rec_t
    EQUIVS['SyntaxMsgs']         = synt_lst_t
    EQUIVS['MethodsMsg']         = methods_msg_t

    _scan_msg_classes(_resolve_forward_refs)
    _scan_msg_classes(_attach_decoders)
    _dispatch_on_sys_exc_types()
    FINALIZED = True
    eprint('\nfinalize_message_classes done\n') if LOG_FINALIZE else None



from agentica_internal.testing.warpc.system_id_tests import *


def test_system_class_ids():
    verify_system_class_ids()


def test_system_function_ids():
    verify_system_function_ids()


def test_system_object_ids():
    verify_system_object_ids()


def test_system_tables_disjoint():
    verify_system_tables_disjoint()
